"""
Test serializers.
"""

from __future__ import absolute_import

import unittest

from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured

from django_elasticsearch_dsl import Document, Index, fields
from django_elasticsearch_dsl_drf_alt.compat import KeywordField, StringField

import pytest

from ..serializers import DocumentSerializer
from .base import BaseRestFrameworkTestCase

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_serializers"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestSerializers",)


@pytest.mark.django_db
class TestSerializers(BaseRestFrameworkTestCase):
    """Test serializers."""

    pytestmark = pytest.mark.django_db

    def _get_user_document(self):
        """Get user document."""
        index = Index("auth_user")

        # See Elasticsearch Indices API reference for available settings
        index.settings(number_of_shards=1, number_of_replicas=1)

        @index.doc_type
        class UserDocument(Document):
            """For testing purposes."""

            id = fields.IntegerField(attr="id")

            username = StringField(
                fields={
                    "raw": KeywordField(),
                    "suggest": fields.CompletionField(),
                }
            )

            first_name = StringField(
                fields={
                    "raw": KeywordField(),
                    "suggest": fields.CompletionField(),
                }
            )

            last_name = StringField(
                fields={
                    "raw": KeywordField(),
                    "suggest": fields.CompletionField(),
                }
            )

            email = StringField(
                fields={
                    "raw": KeywordField(),
                }
            )

            is_staff = fields.BooleanField()

            is_active = fields.BooleanField()

            date_joined = fields.DateField()

            class Django(object):
                model = User  # The model associate with this Document

        return UserDocument

    def _get_user_document_serializer(
        self, user_document, with_document=True, with_exclude=True
    ):
        """Get user document serializer."""

        class UserDocumentSerializer(DocumentSerializer):
            """For testing purposes."""

            class Meta:
                """Meta options."""

                if with_document:
                    document = user_document

                fields = (
                    "id",
                    "username",
                    "first_name",
                    "last_name",
                    "email",
                    "is_staff",
                    "is_active",
                    "date_joined",
                )

                if with_exclude:
                    exclude = (
                        "username",
                        "first_name",
                    )

            def update(self, instance, validated_data):
                return True

        return UserDocumentSerializer

    def _test_serializer_mix_fields_and_exclude(self):
        """Test serializer mix of fields and exclude.

        This method is supposed to return ``ImproperlyConfigured``
        exception.

        """
        user_document = self._get_user_document()
        user_document_serializer = self._get_user_document_serializer(
            user_document, with_document=True, with_exclude=True
        )

        return user_document, user_document_serializer

    def test_serializer_fields_and_exclude(self):
        """Test serializer fields and exclude."""

        self.assertRaises(
            ImproperlyConfigured, self._test_serializer_mix_fields_and_exclude
        )

    def _test_serializer_no_document_specified(self):
        """Test serializer no document specified."""
        user_document = self._get_user_document()
        user_document_serializer = self._get_user_document_serializer(
            user_document, with_document=False, with_exclude=False
        )
        return user_document, user_document_serializer()

    def test_serializer_no_document_specified(self):
        """Test serializer no document specified."""
        self.assertRaises(
            ImproperlyConfigured, self._test_serializer_no_document_specified
        )

    def _test_serializer_document_equals_to_none(self):
        """Test serializer document equals to none."""
        user_document_serializer = self._get_user_document_serializer(
            None, with_document=True, with_exclude=False
        )
        return user_document_serializer()

    def test_serializer_document_equals_to_none(self):
        """Test serializer no document specified."""
        self.assertRaises(
            ImproperlyConfigured, self._test_serializer_document_equals_to_none
        )

    def _test_serializer_meta_set_attr(self):
        """Test serializer set attr."""
        user_document = self._get_user_document()
        user_document_serializer = self._get_user_document_serializer(
            user_document, with_document=True, with_exclude=False
        )
        setattr(user_document_serializer.Meta, "test_name", "test_value")

    def test_serializer_meta_set_attr(self):
        """Test serializer set attr."""
        self.assertRaises(AttributeError, self._test_serializer_meta_set_attr)

    def _test_serializer_meta_del_attr(self):
        """Test serializer del attr."""
        user_document = self._get_user_document()
        user_document_serializer = self._get_user_document_serializer(
            user_document, with_document=True, with_exclude=False
        )
        delattr(user_document_serializer.Meta, "test_name")

    def test_serializer_meta_del_attr(self):
        """Test serializer set attr."""
        self.assertRaises(AttributeError, self._test_serializer_meta_del_attr)


if __name__ == "__main__":
    unittest.main()



"""
Functional suggesters backend.

It's assumed, that fields you're planning to query suggestions for have been
properly indexed using ``fields.CompletionField``.

Example:

    >>> from django_elasticsearch_dsl import Document, Index, fields
    >>>
    >>> from books.models import Publisher
    >>>
    >>> # Name of the Elasticsearch index
    >>> PUBLISHER_INDEX = Index(PUBLISHER_INDEX_NAME)
    >>> # See Elasticsearch Indices API reference for available settings
    >>> PUBLISHER_INDEX.settings(
    >>>     number_of_shards=1,
    >>>     number_of_replicas=1
    >>> )
    >>>
    >>> @PUBLISHER_INDEX.doc_type
    >>> class PublisherDocument(Document):
    >>>     "Publisher Elasticsearch document."
    >>>
    >>>     id = fields.IntegerField(attr='id')
    >>>
    >>>     name = fields.StringField(
    >>>         fields={
    >>>             'raw': fields.StringField(analyzer='keyword'),
    >>>             'suggest': fields.CompletionField(),
    >>>         }
    >>>     )
    >>>
    >>>     info = fields.StringField()
    >>>
    >>>     address = fields.StringField(
    >>>         fields={
    >>>             'raw': fields.StringField(analyzer='keyword')
    >>>         }
    >>>     )
    >>>
    >>>     city = fields.StringField(
    >>>         fields={
    >>>             'raw': fields.StringField(analyzer='keyword'),
    >>>             'suggest': fields.CompletionField(),
    >>>         }
    >>>     )
    >>>
    >>>     state_province = fields.StringField(
    >>>         fields={
    >>>             'raw': fields.StringField(analyzer='keyword'),
    >>>             'suggest': fields.CompletionField(),
    >>>         }
    >>>     )
    >>>
    >>>     country = fields.StringField(
    >>>         fields={
    >>>             'raw': fields.StringField(analyzer='keyword'),
    >>>             'suggest': fields.CompletionField(),
    >>>         }
    >>>     )
    >>>
    >>>     website = fields.StringField()
    >>>
    >>>     class Meta:
    >>>         "Meta options."
    >>>
    >>>         model = Publisher  # The model associate with this Document
"""

try:
    from elasticsearch_dsl.search import AggsProxy
except ImportError:
    from elasticsearch_dsl.search_base import AggsProxy

from django_elasticsearch_dsl_drf_alt.constants import (
    FUNCTIONAL_SUGGESTER_TERM_MATCH,
    FUNCTIONAL_SUGGESTER_PHRASE_MATCH,
    FUNCTIONAL_SUGGESTER_COMPLETION_PREFIX,
    FUNCTIONAL_SUGGESTER_COMPLETION_MATCH,
    ALL_FUNCTIONAL_SUGGESTERS,
)
from django_elasticsearch_dsl_drf_alt.utils import EmptySearch

from rest_framework.exceptions import ValidationError
from rest_framework.filters import BaseFilterBackend

from six import string_types

from ..mixins import FilterBackendMixin

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.suggester." "functional"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("FunctionalSuggesterFilterBackend",)


class FunctionalSuggesterFilterBackend(BaseFilterBackend, FilterBackendMixin):
    """Suggester filter backend for Elasticsearch.

    Suggestion functionality is exclusive. Once you have queried the
    ``FunctionalSuggesterFilterBackend``, the latter will transform your
    current search query into another search query (altered).
    Therefore, always add it as the very last filter backend.

    Example:

        >>> from django_elasticsearch_dsl_drf_alt.constants import (
        >>>     FUNCTIONAL_SUGGESTER_COMPLETION_MATCH,
        >>>     FUNCTIONAL_SUGGESTER_COMPLETION_PREFIX,
        >>>     FUNCTIONAL_SUGGESTER_PHRASE_MATCH,
        >>>     FUNCTIONAL_SUGGESTER_PHRASE_MATCH,
        >>>     FUNCTIONAL_SUGGESTER_TERM_MATCH,
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.filter_backends import (
        >>>     FunctionalSuggesterFilterBackend
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.viewsets import DocumentViewSet
        >>>
        >>> # Local PublisherDocument definition
        >>> from .documents import PublisherDocument
        >>>
        >>> # Local PublisherDocument serializer
        >>> from .serializers import PublisherDocumentSerializer
        >>>
        >>> class PublisherDocumentView(DocumentViewSet):
        >>>
        >>>     document = PublisherDocument
        >>>     serializer_class = PublisherDocumentSerializer
        >>>     filter_backends = [
        >>>         # ...
        >>>         FunctionalSuggesterFilterBackend,
        >>>     ]
        >>>     # Suggester fields
        >>>     functional_suggester_fields = {
        >>>         'name_suggest': {
        >>>             'field': 'name.suggest',
        >>>             'suggesters': [
        >>>                 FUNCTIONAL_SUGGESTER_COMPLETION_MATCH,
        >>>                 FUNCTIONAL_SUGGESTER_COMPLETION_PREFIX,
        >>>             ],
        >>>         },
        >>>         'city_suggest': {
        >>>             'field': 'city.suggest',
        >>>             'suggesters': [
        >>>                 FUNCTIONAL_SUGGESTER_COMPLETION_PREFIX,
        >>>             ],
        >>>         },
        >>>         'state_province_suggest': {
        >>>             'field': 'state_province.suggest',
        >>>             'suggesters': [
        >>>                 FUNCTIONAL_SUGGESTER_COMPLETION_MATCH,
        >>>             ],
        >>>         },
        >>>         'country_suggest': {
        >>>             'field': 'country.suggest',
        >>>             'suggesters': [
        >>>                 FUNCTIONAL_SUGGESTER_COMPLETION_PREFIX,
        >>>             ],
        >>>         },
        >>>     }
    """

    @classmethod
    def prepare_suggester_fields(cls, view):
        """Prepare filter fields.

        :param view:
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Filtering options.
        :rtype: dict
        """
        filter_fields = view.functional_suggester_fields

        for field, options in filter_fields.items():
            if options is None or isinstance(options, string_types):
                filter_fields[field] = {"field": options or field}
            elif "field" not in filter_fields[field]:
                filter_fields[field]["field"] = field

            if "suggesters" not in filter_fields[field]:
                filter_fields[field]["suggesters"] = tuple(ALL_FUNCTIONAL_SUGGESTERS)

        return filter_fields

    # @classmethod
    # def apply_suggester_term(cls, suggester_name, queryset, options, value):
    #     """Apply `term` suggester.
    #
    #     :param suggester_name:
    #     :param queryset: Original queryset.
    #     :param options: Filter options.
    #     :param value: value to filter on.
    #     :type suggester_name: str
    #     :type queryset: elasticsearch_dsl.search.Search
    #     :type options: dict
    #     :type value: str
    #     :return: Modified queryset.
    #     :rtype: elasticsearch_dsl.search.Search
    #     """
    #     return queryset.suggest(
    #         suggester_name,
    #         value,
    #         term={'field': options['field']}
    #     )
    #
    # @classmethod
    # def apply_suggester_phrase(cls,
    #                            suggester_name,
    #                            queryset,
    #                            options,
    #                            value):
    #     """Apply `phrase` suggester.
    #
    #     :param suggester_name:
    #     :param queryset: Original queryset.
    #     :param options: Filter options.
    #     :param value: value to filter on.
    #     :type suggester_name: str
    #     :type queryset: elasticsearch_dsl.search.Search
    #     :type options: dict
    #     :type value: str
    #     :return: Modified queryset.
    #     :rtype: elasticsearch_dsl.search.Search
    #     """
    #     return queryset.suggest(
    #         suggester_name,
    #         value,
    #         phrase={'field': options['field']}
    #     )

    @classmethod
    def apply_query_size(cls, queryset, options):
        """Apply query size.

        :param queryset:
        :param options:
        :return:
        """
        if "size" in options["options"]:
            queryset = queryset.extra(
                from_=options["options"].get("from", 0), size=options["options"]["size"]
            )
        return queryset

    @classmethod
    def apply_suggester_completion_prefix(
        cls, suggester_name, queryset, options, value
    ):
        """Apply `completion` suggester prefix.

        This is effective when used with Keyword fields.

        :param suggester_name:
        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type suggester_name: str
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        queryset = queryset.query("prefix", **{options["field"]: value})
        queryset = cls.apply_query_size(queryset, options)
        return queryset

    @classmethod
    def apply_suggester_completion_match(cls, suggester_name, queryset, options, value):
        """Apply `completion` suggester match.

        This is effective when used with Ngram fields.

        :param suggester_name:
        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type suggester_name: str
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        queryset = queryset.query("match", **{options["field"]: value})
        queryset = cls.apply_query_size(queryset, options)
        return queryset

    def get_suggester_query_params(self, request, view):
        """Get query params to be for suggestions.

        :param request: Django REST framework request.
        :param view: View.
        :type request: rest_framework.request.Request
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Request query params to filter on.
        :rtype: dict
        """
        query_params = request.query_params.copy()

        suggester_query_params = {}
        suggester_fields = self.prepare_suggester_fields(view)
        for query_param in query_params:
            query_param_list = self.split_lookup_filter(query_param, maxsplit=1)
            field_name = query_param_list[0]

            if field_name in suggester_fields:
                suggester_param = None
                if len(query_param_list) > 1:
                    suggester_param = query_param_list[1]

                valid_suggesters = suggester_fields[field_name]["suggesters"]

                suggester_options = {}

                # If we have default suggester given use it as a default and
                # do not require further suffix specification.
                default_suggester = None
                if "default_suggester" in suggester_fields[field_name]:
                    default_suggester = suggester_fields[field_name][
                        "default_suggester"
                    ]

                if "options" in suggester_fields[field_name]:
                    suggester_options = suggester_fields[field_name]["options"]

                if suggester_param is None or suggester_param in valid_suggesters:

                    # If we have default suggester given use it as a default
                    # and do not require further suffix specification.
                    if suggester_param is None and default_suggester is not None:
                        suggester_param = str(default_suggester)

                    values = [
                        __value.strip()
                        for __value in query_params.getlist(query_param)
                        if __value.strip() != ""
                    ]

                    # If specific field given, use that. Otherwise,
                    # fall back to the top level field name.
                    if "serializer_field" in suggester_fields[field_name]:
                        serializer_field = suggester_fields[field_name][
                            "serializer_field"
                        ]
                    else:
                        serializer_field = suggester_fields[field_name].get(
                            "field", field_name
                        )
                        serializer_field = self.extract_field_name(serializer_field)

                    if values:
                        suggester_query_params[query_param] = {
                            "suggester": suggester_param,
                            "values": values,
                            "field": suggester_fields[field_name].get(
                                "field", field_name
                            ),
                            "type": view.mapping,
                            "serializer_field": serializer_field,
                            "options": suggester_options,
                        }
        return suggester_query_params

    def clean_queryset(self, queryset):
        """Clean the queryset.

        - Remove aggregations.
        - Remove highlight.
        - Remove sorting options.

        :param queryset:
        :return:
        """
        queryset.aggs = AggsProxy("")
        queryset._highlight = {}
        queryset._sort = ["_score"]
        queryset._functional_suggest = True
        return queryset

    def serialize_queryset(self, queryset, suggester_name, value, serializer_field):
        """Serialize queryset.

        This shall be done here, since we don't want to delegate it to
        pagination.

        :param queryset:
        :param suggester_name:
        :param value:
        :param serializer_field:
        :return:
        """
        result = queryset.execute().to_dict()
        hits = []
        for hit in result["hits"]["hits"]:
            hit.update({"text": hit["_source"].get(serializer_field)})
            hits.append(hit)

        data = {
            suggester_name: [
                {
                    "text": value,
                    "options": hits,
                    "length": len(value),
                    "offset": 0,
                }
            ],
            "_shards": result["_shards"],
        }
        return data

    def extract_field_name(self, field_name):
        """Extract field name.

        For instance, "name.suggest" or "name.raw" becomes "name".

        :param field_name:
        :type str:
        :return:
        :rtype: str
        """
        return field_name.split(".")[0]

    def filter_queryset(self, request, queryset, view):
        """Filter the queryset.

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        # The ``SuggesterFilterBackend`` filter backend shall be used in
        # the ``suggest`` custom view action/route only. Usages outside of the
        # are ``suggest`` action/route are restricted.
        if view.action != "functional_suggest":
            return queryset

        # Clean the queryset.
        queryset = self.clean_queryset(queryset)

        suggester_query_params = self.get_suggester_query_params(request, view)

        applied = False  # Indicates whether filter has been applied
        picked_suggester_name = None
        picked_value = None
        picked_serializer_field = None

        for suggester_name, options in suggester_query_params.items():
            # We don't have multiple values here.
            for value in options["values"]:
                # `completion` suggester
                if options["suggester"] == FUNCTIONAL_SUGGESTER_COMPLETION_PREFIX:
                    queryset = self.apply_suggester_completion_prefix(
                        suggester_name, queryset, options, value
                    )
                    applied = True
                    picked_suggester_name = suggester_name
                    picked_value = value
                    picked_serializer_field = options["serializer_field"]

                elif options["suggester"] == FUNCTIONAL_SUGGESTER_COMPLETION_MATCH:
                    queryset = self.apply_suggester_completion_match(
                        suggester_name, queryset, options, value
                    )
                    applied = True
                    picked_suggester_name = suggester_name
                    picked_value = value
                    picked_serializer_field = options["serializer_field"]

                # # `term` suggester
                # elif options['suggester'] == SUGGESTER_TERM:
                #     queryset = self.apply_suggester_term(suggester_name,
                #                                          queryset,
                #                                          options,
                #                                          value)
                #
                # # `phrase` suggester
                # elif options['suggester'] == SUGGESTER_PHRASE:
                #     queryset = self.apply_suggester_phrase(suggester_name,
                #                                            queryset,
                #                                            options,
                #                                            value)

        # If no filters have been applied, return empty queryset. This
        # has no affect on other backends, since this only applies to
        # view.action == 'functional_suggest' case.
        if not applied:
            raise ValidationError(detail=None)
            # empty_queryset = EmptySearch()
            # empty_queryset._functional_suggest = True
            # return empty_queryset

        return self.serialize_queryset(
            queryset, picked_suggester_name, picked_value, picked_serializer_field
        )



"""
Fields.
"""

from .common import (
    BooleanField,
    CharField,
    DateField,
    FloatField,
    IntegerField,
    IPAddressField,
)
from .nested_fields import (
    GeoPointField,
    GeoShapeField,
    ListField,
    NestedField,
    ObjectField,
)

__title__ = "django_elasticsearch_dsl_drf_alt.fields"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "BooleanField",
    "CharField",
    "DateField",
    "FloatField",
    "GeoPointField",
    "GeoShapeField",
    "IntegerField",
    "IPAddressField",
    "ListField",
    "NestedField",
    "ObjectField",
)



"""
Utils.
"""

import datetime

try:
    from elasticsearch_dsl.search import AggsProxy
except ImportError:
    from elasticsearch_dsl.search_base import AggsProxy


__title__ = "django_elasticsearch_dsl_drf_alt.utils"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "DictionaryProxy",
    "EmptySearch",
)


class EmptySearch(object):
    """Empty Search."""

    def __init__(self, *args, **kwargs):
        self.aggs = AggsProxy("")
        self._highlight = {}
        self._sort = []
        self.total = 0

    def __len__(self):
        return 0

    def __iter__(self):
        return iter([])

    def __getitem__(self, *args, **kwargs):
        return self

    def highlight(self, *args, **kwargs):
        return self

    def sort(self, *args, **kwargs):
        return self

    @property
    def hits(self):
        return self

    def execute(self, *args, **kwargs):
        return self

    def to_dict(self, *args, **kwargs):
        return {}


class DictionaryProxy(object):
    """Dictionary proxy."""

    def __init__(self, mapping, meta=None):
        self.__mapping = mapping
        self.meta = meta

    def __getattr__(self, item):
        if item == "meta":
            return self.meta
        val = self.__mapping.get(item, None)
        if isinstance(val, datetime.datetime) and not val.tzinfo:
            val = val.date()
        return val

    def to_dict(self):
        """To dict.

        :return:
        """
        return self.__mapping



from elasticsearch_dsl.query import Q

from .base import BaseSearchQueryBackend

__title__ = (
    "django_elasticsearch_dsl_drf_alt.filter_backends.search." "query_backends.match"
)
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("MatchQueryBackend",)


class MatchQueryBackend(BaseSearchQueryBackend):
    """Match query backend."""

    query_type = "match"

    @classmethod
    def construct_search(cls, request, view, search_backend):
        """Construct search.

        :param request:
        :param view:
        :param search_backend:
        :return:
        """
        query_params = search_backend.get_search_query_params(request)
        __queries = []
        for search_term in query_params:
            __values = search_backend.split_lookup_name(search_term, 1)
            __len_values = len(__values)
            if __len_values > 1:
                field, value = __values
                if field in view.search_fields:
                    # Initial kwargs for the match query
                    field_kwargs = {field: {"query": value}}
                    # In case if we deal with structure 2
                    if isinstance(view.search_fields, dict):
                        extra_field_kwargs = view.search_fields[field]
                        if extra_field_kwargs:
                            field_kwargs[field].update(extra_field_kwargs)
                    # The match query
                    __queries.append(Q(cls.query_type, **field_kwargs))
            else:
                for field in view.search_fields:
                    # Initial kwargs for the match query
                    field_kwargs = {field: {"query": search_term}}

                    # In case if we deal with structure 2
                    if isinstance(view.search_fields, dict):
                        extra_field_kwargs = view.search_fields[field]
                        if extra_field_kwargs:
                            field_kwargs[field].update(extra_field_kwargs)

                    # The match query
                    __queries.append(Q(cls.query_type, **field_kwargs))
        return __queries



"""Multi match search filter backend."""

import logging

from ...constants import MATCHING_OPTION_MUST
from .base import BaseSearchFilterBackend
from .query_backends import (
    MultiMatchQueryBackend,
)

LOGGER = logging.getLogger(__name__)


__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.search.multi_match"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("MultiMatchSearchFilterBackend",)


class MultiMatchSearchFilterBackend(BaseSearchFilterBackend):
    """Multi match search filter backend."""

    search_param = "search_multi_match"

    matching = MATCHING_OPTION_MUST

    query_backends = [
        MultiMatchQueryBackend,
    ]



"""
Suggester filtering backends.
"""

from .functional import FunctionalSuggesterFilterBackend
from .native import SuggesterFilterBackend

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.suggester"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "SuggesterFilterBackend",
    "FunctionalSuggesterFilterBackend",
)



"""
Test faceted search backend.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

from books import constants
import factories

from .base import BaseRestFrameworkTestCase

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_faceted_search"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestFacetedFilteredSearch",)


@pytest.mark.django_db
class TestFacetedFilteredSearch(BaseRestFrameworkTestCase):
    """Test faceted filtered search."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        super(TestFacetedFilteredSearch, cls).setUpClass()

        # published with a known title
        cls.prefixed = factories.BookFactory.create(
            **{
                "title": "My first book",
                "state": constants.BOOK_PUBLISHING_STATUS_PUBLISHED,
            }
        )

        cls.published_count = 11
        cls.published = factories.BookFactory.create_batch(
            cls.published_count - 1,
            **{
                "state": constants.BOOK_PUBLISHING_STATUS_PUBLISHED,
            }
        )

        cls.not_published_count = 10
        cls.not_published = factories.BookFactory.create_batch(
            cls.not_published_count,
            **{
                "state": constants.BOOK_PUBLISHING_STATUS_NOT_PUBLISHED,
            }
        )

        cls.all_count = cls.published_count + cls.not_published_count

        cls.sleep()
        call_command("search_index", "--rebuild", "-f")

    def test_list_results_no_facets(self):
        """List results without facets."""
        self.authenticate()
        url = reverse("bookdocument_faceted_filtered-list")

        # Make request
        no_args_response = self.client.get(url)
        self.assertEqual(no_args_response.status_code, status.HTTP_200_OK)

        # Should contain `self.all_count` results
        self.assertEqual(len(no_args_response.data["results"]), self.all_count)

        # Should contain no facets
        print(no_args_response.data["facets"])
        self.assertEqual(no_args_response.data["facets"], {})

    def test_list_results_facet_no_filter(self):
        self.authenticate()
        url = reverse("bookdocument_faceted_filtered-list")

        response = self.client.get(url + "?facet=state")
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # Should contain `self.all_count` results
        self.assertEqual(len(response.data["results"]), self.all_count)

        # Should contain 1 facet
        self.assertEqual(len(response.data["facets"]), 1)
        self.assertEqual(
            response.data["facets"]["_filter_state"]["state"]["buckets"],
            [
                {"doc_count": self.published_count, "key": "published"},
                {"doc_count": self.not_published_count, "key": "not_published"},
            ],
        )

    def test_list_results_facet_and_filter(self):
        self.authenticate()
        url = reverse("bookdocument_faceted_filtered-list")

        response = self.client.get(
            url
            + "?facet=state&state={}".format(constants.BOOK_PUBLISHING_STATUS_PUBLISHED)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # Should contain all published books
        self.assertEqual(len(response.data["results"]), self.published_count)

        # Should contain 1 facet
        self.assertEqual(len(response.data["facets"]), 1)
        self.assertEqual(
            response.data["facets"]["_filter_state"]["state"]["buckets"],
            [
                {"doc_count": self.published_count, "key": "published"},
                {"doc_count": self.not_published_count, "key": "not_published"},
            ],
        )

    def test_list_results_facet_and_filter_on_non_faceted_field(self):
        self.authenticate()
        url = reverse("bookdocument_faceted_filtered-list")

        response = self.client.get(
            url
            + "?facet=state&state={}&title={}".format(
                constants.BOOK_PUBLISHING_STATUS_PUBLISHED,
                "My first book",
            )
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # Should contain 1 result
        self.assertEqual(len(response.data["results"]), 1)

        # Should contain 1 facet which matches only the single book, since that is applied as a pre-filter
        self.assertEqual(len(response.data["facets"]), 1)
        self.assertEqual(
            response.data["facets"]["_filter_state"]["state"]["buckets"],
            [{"doc_count": 1, "key": "published"}],
        )

    def test_list_results_facet_and_filter_only_on_non_faceted_field(self):
        self.authenticate()
        url = reverse("bookdocument_faceted_filtered-list")

        response = self.client.get(
            url + "?facet=state&title={}".format("My first book")
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # Should contain 1 result
        self.assertEqual(len(response.data["results"]), 1)

        # Should contain 1 facet which matches only the single book, since that is applied as a pre-filter
        self.assertEqual(len(response.data["facets"]), 1)
        self.assertEqual(
            response.data["facets"]["_filter_state"]["state"]["buckets"],
            [{"doc_count": 1, "key": "published"}],
        )


if __name__ == "__main__":
    unittest.main()



"""
Constants module. Contains Elasticsearch constants, lookup constants,
functional constants, suggesters, etc.
"""

__title__ = "django_elasticsearch_dsl_drf_alt.constants"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "ALL_FUNCTIONAL_SUGGESTERS",
    "ALL_GEO_SPATIAL_LOOKUP_FILTERS_AND_QUERIES",
    "ALL_LOOKUP_FILTERS_AND_QUERIES",
    "ALL_SUGGESTERS",
    "DEFAULT_MATCHING_OPTION",
    "EXTENDED_NUMBER_LOOKUP_FILTERS",
    "EXTENDED_STRING_LOOKUP_FILTERS",
    "FALSE_VALUES",
    "FUNCTIONAL_SUGGESTER_COMPLETION_MATCH",
    "FUNCTIONAL_SUGGESTER_COMPLETION_PREFIX",
    "FUNCTIONAL_SUGGESTER_PHRASE_MATCH",
    "FUNCTIONAL_SUGGESTER_TERM_MATCH",
    "GEO_DISTANCE_ORDERING_PARAM",
    "LOOKUP_FILTER_EXISTS",
    "LOOKUP_FILTER_GEO_BOUNDING_BOX",
    "LOOKUP_FILTER_GEO_DISTANCE",
    "LOOKUP_FILTER_GEO_POLYGON",
    "LOOKUP_FILTER_GEO_SHAPE",
    "LOOKUP_FILTER_PREFIX",
    "LOOKUP_FILTER_RANGE",
    "LOOKUP_FILTER_REGEXP",
    "LOOKUP_FILTER_TERM",
    "LOOKUP_FILTER_TERMS",
    "LOOKUP_FILTER_WILDCARD",
    "LOOKUP_QUERY_CONTAINS",
    "LOOKUP_QUERY_ENDSWITH",
    "LOOKUP_QUERY_EXCLUDE",
    "LOOKUP_QUERY_GT",
    "LOOKUP_QUERY_GTE",
    "LOOKUP_QUERY_IN",
    "LOOKUP_QUERY_ISNULL",
    "LOOKUP_QUERY_LT",
    "LOOKUP_QUERY_LTE",
    "LOOKUP_QUERY_STARTSWITH",
    "MATCHING_OPTION_MUST",
    "MATCHING_OPTION_SHOULD",
    "MATCHING_OPTIONS",
    "NUMBER_LOOKUP_FILTERS",
    "SEARCH_QUERY_PARAM",
    "SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE",
    "SEPARATOR_LOOKUP_COMPLEX_VALUE",
    "SEPARATOR_LOOKUP_FILTER",
    "SEPARATOR_LOOKUP_NAME",
    "STRING_LOOKUP_FILTERS",
    "SUGGESTER_COMPLETION",
    "SUGGESTER_PHRASE",
    "SUGGESTER_TERM",
    "TRUE_VALUES",
)

# ****************************************************************************
# ****************************** True / False ********************************
# ****************************************************************************

# As mentioned in official documentation
# https://www.elastic.co/guide/en/elasticsearch/reference/current/boolean.html
# False values: false, "false", "off", "no", "0", "" (empty string), 0, 0.0
# True values: Anything that isn't false.
# Elasticsearch 5.1 currently accepts the above mentioned values during index
# time. Searching a boolean field using these pseudo-boolean values is
# deprecated. You should be using "true" or "false" instead.
# As of 5.3.0, usage of any value other than false, "false", true and "true"
# is deprecated.
# For the time being we'are supporting all values, but you are not recommended
# to use anything except: true, "true", false, "false".

# True values
TRUE_VALUES = (
    "true",
    '"true"',
    "1",  # To be deprecated
)

# False values
FALSE_VALUES = (
    "false",
    '"false"',
    '"off"',  # To be deprecated
    '"no"',  # To be deprecated
    '"0"',  # To be deprecated
    '""',  # To be deprecated
    "",  # To be deprecated
    "0",  # To be deprecated
    "0.0",  # To be deprecated
)

# ****************************************************************************
# ****************************** Lookup related ******************************
# ****************************************************************************

# Lookup separator
SEPARATOR_LOOKUP_FILTER = "__"

# Lookup filter value separator. To be used for `terms` and `range` filters
# lookups.
SEPARATOR_LOOKUP_NAME = ":"  # Used to be SEPARATOR_LOOKUP_VALUE

# Lookup filter value complex separator. To be used with geo-spatial features.
SEPARATOR_LOOKUP_COMPLEX_VALUE = "__"

SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE = ","

# Search query param
SEARCH_QUERY_PARAM = "q"

# Geo distance ordering param
GEO_DISTANCE_ORDERING_PARAM = "ordering"

# ****************************************************************************
# ************************ Native lookup filters/queries *********************
# ****************************************************************************
# Lookup filters and queries that are native to Elasticsearch
# https://www.elastic.co/guide/en/elasticsearch/reference/current/
# term-level-queries.html

# The `term` filter. Accepts a single value.
# Example: {"filter": {"term": {"tags": "children"}}}
# Example: http://localhost:8000/api/articles/?tags=children
# https://www.elastic.co/guide/en/elasticsearch/reference/current/
# query-dsl-term-query.html
LOOKUP_FILTER_TERM = "term"

# The `terms` filter. Should accept multiple values, separated by
# `SEPARATOR_LOOKUP_COMPLEX_VALUE`.
# Example: {"filter": {"terms": {"tags": ["python", "children"]}}}
# Example: http://localhost:8000/api/articles/?tags__terms=children__python
# https://www.elastic.co/guide/en/elasticsearch/reference/current/
# query-dsl-terms-query.html
LOOKUP_FILTER_TERMS = "terms"

# The `range` filter. Accepts a pair of values separated by
# `SEPARATOR_LOOKUP_COMPLEX_VALUE`.
# Example: {"query": {"range": {"age": {"gte": "16",
#                                       "lte": "67",
#                                       "boost": 2.0}}}}
# Example: http://localhost:8000/api/users/?age__range=16__67__2.0
# Example: {"query": {"range": {"age": {"gte": "16", "lte": "67"}}}}
# Example: http://localhost:8000/api/users/?age__range=16__67
# Example: {"query": {"range": {"age": {"gte": "16"}}}}
# Example: http://localhost:8000/api/users/?age__range=16
# https://www.elastic.co/guide/en/elasticsearch/reference/current/
# query-dsl-range-query.html
LOOKUP_FILTER_RANGE = "range"

# Returns documents that have at least one non-null value in the original
# field.
# Example: {"query": {"exists": {"field": "tags"}}}
# Example: http://localhost:8000/api/articles/?tags__exists=true
# https://www.elastic.co/guide/en/elasticsearch/reference/current/
# query-dsl-exists-query.html
LOOKUP_FILTER_EXISTS = "exists"

# The `prefix` filter. Accepts a single value.
# Example: {"filter": {"prefix": {"tags": "bio"}}}
# Example: http://localhost:8000/api/articles/?tags__prefix=bio
# https://www.elastic.co/guide/en/elasticsearch/reference/current/
# query-dsl-prefix-query.html
LOOKUP_FILTER_PREFIX = "prefix"

# Supported wildcards are `*`, which matches any character sequence (including
# the empty one), and `?`, which matches any single character. Note that this
# query can be slow, as it needs to iterate over many terms. In order to
# prevent extremely slow wildcard queries, a wildcard term should not start
# with one of the wildcards `*` or `?`.
# Example: {"filter": {"wildcard": {"tags": "child*"}}}
# Example: http://localhost:8000/api/articles/?tags__wildcard=child*
# https://www.elastic.co/guide/en/elasticsearch/reference/current/
# query-dsl-wildcard-query.html
LOOKUP_FILTER_WILDCARD = "wildcard"

# TODO: Implement
# The regexp query allows you to use regular expression term queries. See
# Regular expression syntax for details of the supported regular expression
# language. The "term queries" in that first sentence means that Elasticsearch
# will apply the regexp to the terms produced by the tokenizer for that field,
# and not to the original text of the field.
# Note: The performance of a regexp query heavily depends on the regular
# expression chosen. Matching everything like `.*` is very slow as well as
# using lookaround regular expressions. If possible, you should try to use a
# long prefix before your regular expression starts. Wildcard matchers
# like `.*?+` will mostly lower performance.
LOOKUP_FILTER_REGEXP = "regexp"
# Example: {"query": {"regexp": {"tags": "ch.*en"}}}
# Example: http://localhost:8000/api/articles/?tags__regexp=ch.*en
# https://www.elastic.co/guide/en/elasticsearch/reference/current/
# query-dsl-regexp-query.html

# TODO: Implement
# The fuzzy query uses similarity based on Levenshtein edit distance.
LOOKUP_FILTER_FUZZY = "fuzzy"

# TODO: Implement
# Filters documents matching the provided document / mapping type.
# https://www.elastic.co/guide/en/elasticsearch/reference/current/
# query-dsl-type-query.html
LOOKUP_FILTER_TYPE = "type"

# ****************************************************************************
# ******************* Native geo-spatial lookup filters/queries **************
# ****************************************************************************

# Draws a circle around the specified location and finds all documents that
# have a geo-point within that circle.
#
# The `geo_distance` filter. Accepts three values (distance__lat__lon)
# separated by `SEPARATOR_LOOKUP_COMPLEX_VALUE`.
# https://www.elastic.co/guide/en/elasticsearch/guide/current/geo-distance.html
#
# Example:
#
# {
#     "query": {
#         "bool" : {
#             "must" : {
#                 "match_all" : {}
#             },
#             "filter" : {
#                 "geo_distance" : {
#                     "distance" : "200km",
#                     "pin.location" : {
#                         "lat" : 40,
#                         "lon" : -70
#                     }
#                 }
#             }
#         }
#     }
#     }
#
# Example: http://localhost:8000
# /api/articles/?location__geo_distance=2km__43.53455243__-12.2344243
LOOKUP_FILTER_GEO_DISTANCE = "geo_distance"

# Geo Polygon Query
#
# A query allowing to include hits that only fall within a polygon of points.
# Example:
#
# {
#     "query": {
#         "bool" : {
#             "must" : {
#                 "match_all" : {}
#             },
#             "filter" : {
#                 "geo_polygon" : {
#                     "person.location" : {
#                         "points" : [
#                         {"lat" : 40, "lon" : -70},
#                         {"lat" : 30, "lon" : -80},
#                         {"lat" : 20, "lon" : -90}
#                         ]
#                     }
#                 }
#             }
#         }
#     }
# }
#
# Query options:
#
# - _name: Optional name field to identify the filter
# - validation_method: Set to IGNORE_MALFORMED to accept geo points with
#   invalid latitude or longitude, COERCE to try and infer correct latitude or
#   longitude, or STRICT (default is STRICT).
#
# Example: http://localhost:8000
# /api/articles/?location__geo_polygon=40,-70__30,-80__20,-90
LOOKUP_FILTER_GEO_POLYGON = "geo_polygon"

# Geo Bounding Box Query
# A query allowing to filter hits based on a point location using a bounding
# box. Assuming the following indexed document:
#
# Example:
#
# {
#     "query": {
#         "bool" : {
#             "must" : {
#                 "match_all" : {}
#             },
#             "filter" : {
#                 "geo_bounding_box" : {
#                     "location" : {
#                         "top_left" : {
#                             "lat" : 40.73,
#                             "lon" : -74.1
#                         },
#                         "bottom_right" : {
#                             "lat" : 40.01,
#                             "lon" : -71.12
#                         }
#                     }
#                 }
#             }
#         }
#     }
# }
#
# Query options:
#
# - _name: Optional name field to identify the filter
# - validation_method: Set to IGNORE_MALFORMED to accept geo points with
#   invalid latitude or longitude, COERCE to try and infer correct latitude or
#   longitude, or STRICT (default is STRICT).
# - type: Set to one of indexed or memory to defines whether this filter will
#   be executed in memory or indexed. See Type below for further details.
#   Default is memory.
#
# Example: http://localhost:8000
# /api/articles/?location__geo_bounding_box=40.73,-74.1__40.01,-71.12
LOOKUP_FILTER_GEO_BOUNDING_BOX = "geo_bounding_box"

# Geo Shape Query
# Finds documents with:
#   - geo-shapes which either intersect, are contained by, or do not intersect with the specified geo-shape
#   - geo-points which intersect the specified geo-shape
#
# Example:
# {
#     "query": {
#         "bool" : {
#             "must" : {
#                 "match_all" : {}
#             },
#             "filter": {
#                 "geo_shape": {
#                     "location": {
#                         "shape": {
#                             "type": "circle",
#                             "coordinates": [48.9864453, 6.37977],
#                             "radius": "20km"
#                         },
#                         "relation": "intersects"
#                     }
#                 }
#             }
#         }
#     }
# }
#
# Query options:
#
# - type: Shape type (envelope, circle, polygon, ...)
# - relation: Spatial relation operator (intersects, disjoint, within, ...)
# - radius: in case of circle type, represents circle radius
#
# Example: http://localhost:8000
# /api/articles/?location__geo_shape=48.9864453,6.37977__relation,intersects__type,circle__radius,20km
LOOKUP_FILTER_GEO_SHAPE = "geo_shape"

# ****************************************************************************
# ************************ Functional filters/queries ************************
# ****************************************************************************
# Lookup queries that are not native to Elasticsearch, but rather handy/easy
# to use. Inspired by Django's ORM lookups and other sources.
# https://docs.djangoproject.com/en/1.11/ref/models/querysets/#id4

# A single value
# http://localhost:8000/api/articles/?state__contains=lish
LOOKUP_QUERY_CONTAINS = "contains"

# Multiple values.
# Example: http://localhost:8000/api/articles/?tags__in=children__python
LOOKUP_QUERY_IN = "in"

# A single value
# Example: http://localhost:8000/api/articles/?id__gt=1
# Example: http://localhost:8000/api/articles/?id__gt=1__2.0
# https://www.elastic.co/guide/en/elasticsearch/reference/1.6/
# query-dsl-range-filter.html
LOOKUP_QUERY_GT = "gt"

# A single value
# Example: http://localhost:8000/api/articles/?id__gte=1
# Example: http://localhost:8000/api/articles/?id__gte=1__2.0
# https://www.elastic.co/guide/en/elasticsearch/reference/1.6/
# query-dsl-range-filter.html
LOOKUP_QUERY_GTE = "gte"

# A single value
# Example: http://localhost:8000/api/articles/?id__lt=1
# Example: http://localhost:8000/api/articles/?id__lt=1__2.0
# https://www.elastic.co/guide/en/elasticsearch/reference/1.6/
# query-dsl-range-filter.html
LOOKUP_QUERY_LT = "lt"

# A single value
# Example: http://localhost:8000/api/articles/?id__lte=1
# Example: http://localhost:8000/api/articles/?id__lte=1__2.0
# https://www.elastic.co/guide/en/elasticsearch/reference/1.6/
# query-dsl-range-filter.html
LOOKUP_QUERY_LTE = "lte"

# A single value. Alias of `prefix`.
# Example: http://localhost:8000/api/articles/?tags__startswith=chil
LOOKUP_QUERY_STARTSWITH = "startswith"

# A single value
# Example: http://localhost:8000/api/articles/?tags__endswith=dren
# Example: http://localhost:8000/api/articles/?state__endswith=lished
LOOKUP_QUERY_ENDSWITH = "endswith"

# A single value
# Example: http://localhost:8000/api/articles/?tags__isnull=true
LOOKUP_QUERY_ISNULL = "isnull"

# Multiple values.
# Example: http://localhost:8000/api/articles/?tags__exclude=children
# Example: http://localhost:8000/api/articles/?tags__exclude=children__python
LOOKUP_QUERY_EXCLUDE = "exclude"

# ****************************************************************************
# *************************** Suggestions filters ****************************
# ****************************************************************************
# http://elasticsearch-dsl.readthedocs.io/en/latest/search_dsl.html#suggestions

# The `term` suggester
# http://127.0.0.1:8000/search/books/?title_suggest__term=Lore
SUGGESTER_TERM = "term"

# The `phrase` suggester
# http://127.0.0.1:8000/search/books/?title_suggest__phrase=Lorem
SUGGESTER_PHRASE = "phrase"

# The `completion` suggester
# http://127.0.0.1:8000/search/books/?title_suggest__completion=Lore
SUGGESTER_COMPLETION = "completion"

# ****************************************************************************
# ********************** Functional suggestions filters **********************
# ****************************************************************************
# http://elasticsearch-dsl.readthedocs.io/en/latest/search_dsl.html#suggestions

# The `term` suggester
# http://127.0.0.1:8000/search/books/?title_suggest__term=Lore
FUNCTIONAL_SUGGESTER_TERM_MATCH = "term_match"

# The `phrase` suggester
# http://127.0.0.1:8000/search/books/?title_suggest__phrase=Lorem
FUNCTIONAL_SUGGESTER_PHRASE_MATCH = "phrase_match"

# The `completion` suggester
# http://127.0.0.1:8000/search/books/?title_suggest__completion_match=Lore
FUNCTIONAL_SUGGESTER_COMPLETION_MATCH = "completion_match"

# The `completion` suggester
# http://127.0.0.1:8000/search/books/?title_suggest__completion_prefix=Lore
FUNCTIONAL_SUGGESTER_COMPLETION_PREFIX = "completion_prefix"

# ****************************************************************************
# ******************************* Combinations *******************************
# ****************************************************************************
# Combinations of multiple constants.

# All lookup filters and queries
ALL_LOOKUP_FILTERS_AND_QUERIES = (
    # Native
    LOOKUP_FILTER_TERM,
    LOOKUP_FILTER_TERMS,
    LOOKUP_FILTER_RANGE,
    LOOKUP_FILTER_EXISTS,
    LOOKUP_FILTER_PREFIX,
    LOOKUP_FILTER_WILDCARD,
    LOOKUP_FILTER_REGEXP,
    # LOOKUP_FILTER_FUZZY,
    # LOOKUP_FILTER_TYPE,
    # Functional
    LOOKUP_QUERY_CONTAINS,
    LOOKUP_QUERY_IN,
    LOOKUP_QUERY_GT,
    LOOKUP_QUERY_GTE,
    LOOKUP_QUERY_LT,
    LOOKUP_QUERY_LTE,
    LOOKUP_QUERY_STARTSWITH,
    LOOKUP_QUERY_ENDSWITH,
    LOOKUP_QUERY_ISNULL,
    LOOKUP_QUERY_EXCLUDE,
)

ALL_SUGGESTERS = (
    SUGGESTER_TERM,
    SUGGESTER_PHRASE,
    SUGGESTER_COMPLETION,
)

ALL_FUNCTIONAL_SUGGESTERS = (
    FUNCTIONAL_SUGGESTER_TERM_MATCH,
    FUNCTIONAL_SUGGESTER_PHRASE_MATCH,
    FUNCTIONAL_SUGGESTER_COMPLETION_MATCH,
    FUNCTIONAL_SUGGESTER_COMPLETION_PREFIX,
)

ALL_GEO_SPATIAL_LOOKUP_FILTERS_AND_QUERIES = (
    LOOKUP_FILTER_GEO_BOUNDING_BOX,
    LOOKUP_FILTER_GEO_DISTANCE,
    LOOKUP_FILTER_GEO_POLYGON,
)

STRING_LOOKUP_FILTERS = [
    LOOKUP_FILTER_TERM,
    LOOKUP_FILTER_TERMS,
    LOOKUP_FILTER_PREFIX,
    LOOKUP_FILTER_WILDCARD,
    LOOKUP_FILTER_REGEXP,
    LOOKUP_QUERY_CONTAINS,
    LOOKUP_QUERY_IN,
    LOOKUP_QUERY_STARTSWITH,
    LOOKUP_QUERY_ENDSWITH,
    LOOKUP_QUERY_EXCLUDE,
]

EXTENDED_STRING_LOOKUP_FILTERS = STRING_LOOKUP_FILTERS + [
    LOOKUP_FILTER_EXISTS,
    LOOKUP_QUERY_ISNULL,
]

NUMBER_LOOKUP_FILTERS = [
    LOOKUP_FILTER_RANGE,
    LOOKUP_QUERY_GT,
    LOOKUP_QUERY_GTE,
    LOOKUP_QUERY_IN,
    LOOKUP_QUERY_LT,
    LOOKUP_QUERY_LTE,
    LOOKUP_QUERY_EXCLUDE,
]

EXTENDED_NUMBER_LOOKUP_FILTERS = NUMBER_LOOKUP_FILTERS + [
    LOOKUP_FILTER_EXISTS,
    LOOKUP_QUERY_ISNULL,
]

MATCHING_OPTION_MUST = "must"
MATCHING_OPTION_SHOULD = "should"
MATCHING_OPTIONS = (
    MATCHING_OPTION_MUST,
    MATCHING_OPTION_SHOULD,
)
DEFAULT_MATCHING_OPTION = MATCHING_OPTION_SHOULD



"""
Search query backends.
"""

from .base import BaseSearchQueryBackend
from .match import MatchQueryBackend
from .match_phrase import MatchPhraseQueryBackend
from .match_phrase_prefix import MatchPhrasePrefixQueryBackend
from .multi_match import MultiMatchQueryBackend
from .nested import NestedQueryBackend
from .simple_query_string import SimpleQueryStringQueryBackend

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.search." "query_backends"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "BaseSearchQueryBackend",
    "MatchQueryBackend",
    "MatchPhraseQueryBackend",
    "MatchPhrasePrefixQueryBackend",
    "MultiMatchQueryBackend",
    "NestedQueryBackend",
    "SimpleQueryStringQueryBackend",
)



"""
Common fields.
"""

from rest_framework import serializers
from .helpers import to_representation

__title__ = "django_elasticsearch_dsl_drf_alt.fields.nested_fields"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "BooleanField",
    "CharField",
    "DateField",
    "FloatField",
    "IntegerField",
    "IPAddressField",
)


class BooleanField(serializers.BooleanField):
    """Object field."""

    def get_value(self, dictionary):
        """Get value."""
        value = super(BooleanField, self).get_value(dictionary)
        return to_representation(value)

    def to_representation(self, value):
        """To representation."""
        return to_representation(value)


class CharField(serializers.CharField):
    """Object field."""

    def get_value(self, dictionary):
        """Get value."""
        value = super(CharField, self).get_value(dictionary)
        return to_representation(value)

    def to_representation(self, value):
        """To representation."""
        return to_representation(value)


class DateField(serializers.DateField):
    """Object field."""

    def get_value(self, dictionary):
        """Get value."""
        value = super(DateField, self).get_value(dictionary)
        return to_representation(value)

    def to_representation(self, value):
        """To representation."""
        return to_representation(value)


class FloatField(serializers.FloatField):
    """Object field."""

    def get_value(self, dictionary):
        """Get value."""
        value = super(FloatField, self).get_value(dictionary)
        return to_representation(value)

    def to_representation(self, value):
        """To representation."""
        return to_representation(value)


class IntegerField(serializers.IntegerField):
    """Object field."""

    def get_value(self, dictionary):
        """Get value."""
        value = super(IntegerField, self).get_value(dictionary)
        return to_representation(value)

    def to_representation(self, value):
        """To representation."""
        return to_representation(value)


class IPAddressField(serializers.IPAddressField):
    """Object field."""

    def get_value(self, dictionary):
        """Get value."""
        value = super(IPAddressField, self).get_value(dictionary)
        return to_representation(value)

    def to_representation(self, value):
        """To representation."""
        return to_representation(value)



"""
The ``post_filter`` filtering backend.
"""

from django_elasticsearch_dsl import fields

from six import string_types

from ...compat import coreapi
from ...compat import coreschema
from ...constants import ALL_LOOKUP_FILTERS_AND_QUERIES

from .common import FilteringFilterBackend

__title__ = (
    "django_elasticsearch_dsl_drf_alt.filter_backends.filtering." "post_filter.common"
)
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("PostFilterFilteringFilterBackend",)


class PostFilterFilteringFilterBackend(FilteringFilterBackend):
    """The ``post_filter`` filtering filter backend for Elasticsearch.

    Example:

        >>> from django_elasticsearch_dsl_drf_alt.constants import (
        >>>     LOOKUP_FILTER_PREFIX,
        >>>     LOOKUP_FILTER_WILDCARD,
        >>>     LOOKUP_QUERY_EXCLUDE,
        >>>     LOOKUP_QUERY_ISNULL,
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.filter_backends import (
        >>>     PostFilterFilteringFilterBackend
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.viewsets import (
        >>>     BaseDocumentViewSet,
        >>> )
        >>>
        >>> # Local article document definition
        >>> from .documents import ArticleDocument
        >>>
        >>> # Local article document serializer
        >>> from .serializers import ArticleDocumentSerializer
        >>>
        >>> class ArticleDocumentView(BaseDocumentViewSet):
        >>>
        >>>     document = ArticleDocument
        >>>     serializer_class = ArticleDocumentSerializer
        >>>     filter_backends = [PostFilterFilteringFilterBackend,]
        >>>     post_filter_fields = {
        >>>         'title': 'title.raw',
        >>>         'state': {
        >>>             'field': 'state.raw',
        >>>             'lookups': [
        >>>                 LOOKUP_FILTER_PREFIX,
        >>>                 LOOKUP_FILTER_WILDCARD,
        >>>                 LOOKUP_QUERY_EXCLUDE,
        >>>                 LOOKUP_QUERY_ISNULL,
        >>>             ],
        >>>         }
        >>> }
    """

    @classmethod
    def prepare_filter_fields(cls, view):
        """Prepare filter fields.

        :param view:
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Filtering options.
        :rtype: dict
        """
        filter_fields = view.post_filter_fields

        for field, options in filter_fields.items():
            if options is None or isinstance(options, string_types):
                filter_fields[field] = {"field": options or field}
            elif "field" not in filter_fields[field]:
                filter_fields[field]["field"] = field

            if "lookups" not in filter_fields[field]:
                filter_fields[field]["lookups"] = tuple(ALL_LOOKUP_FILTERS_AND_QUERIES)

        return filter_fields

    @classmethod
    def apply_filter(cls, queryset, options=None, args=None, kwargs=None):
        """Apply filter.

        :param queryset:
        :param options:
        :param args:
        :param kwargs:
        :return:
        """
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        return queryset.post_filter(*args, **kwargs)

    @classmethod
    def apply_query(cls, queryset, options=None, args=None, kwargs=None):
        """Apply query.

        :param queryset:
        :param options:
        :param args:
        :param kwargs:
        :return:
        """
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        return queryset.post_filter(*args, **kwargs)

    def get_coreschema_field(self, field):
        if isinstance(field, fields.IntegerField):
            field_cls = coreschema.Number
        else:
            field_cls = coreschema.String
        return field_cls()

    def get_schema_fields(self, view):
        assert coreapi is not None, (
            "coreapi must be installed to " "use `get_schema_fields()`"
        )
        assert coreschema is not None, (
            "coreschema must be installed to " "use `get_schema_fields()`"
        )
        filter_fields = getattr(view, "post_filter_fields", None)
        document = getattr(view, "document", None)

        return (
            []
            if not filter_fields
            else [
                coreapi.Field(
                    name=field_name,
                    required=False,
                    location="query",
                    schema=self.get_coreschema_field(document._fields.get(field_name)),
                )
                for field_name in filter_fields
            ]
        )



# -*- coding: utf-8 -*-

"""
Base ViewSets.
"""
from __future__ import absolute_import, unicode_literals

import copy

from django.http import Http404
from django.core.exceptions import ImproperlyConfigured

from elasticsearch_dsl import Search
from elasticsearch_dsl.connections import connections
from elasticsearch_dsl.query import MoreLikeThis

from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.viewsets import ReadOnlyModelViewSet

from .pagination import PageNumberPagination
from .utils import DictionaryProxy
from .versions import ELASTICSEARCH_GTE_7_0

__title__ = "django_elasticsearch_dsl_drf_alt.viewsets"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "BaseDocumentViewSet",
    "DocumentViewSet",
    "FunctionalSuggestMixin",
    "MoreLikeThisMixin",
    "SuggestMixin",
)


class SuggestMixin(object):
    """Suggest mixin."""

    @action(detail=False)
    def suggest(self, request):
        """Suggest functionality."""
        queryset = self.filter_queryset(self.get_queryset())
        is_suggest = getattr(queryset, "_suggest", False)
        if not is_suggest:
            return Response(status=status.HTTP_400_BAD_REQUEST)

        page = self.paginate_queryset(queryset)
        return Response(page)


class FunctionalSuggestMixin(object):
    """Functional suggest mixin."""

    @action(detail=False)
    def functional_suggest(self, request):
        """Functional suggest functionality.

        :param request:
        :return:
        """
        # TODO: leave the following check or remove?
        if "view" in request.parser_context:
            view = request.parser_context["view"]
            filter_backend_names = [__b.__name__ for __b in view.filter_backends]
            if "FunctionalSuggesterFilterBackend" not in filter_backend_names:
                raise ImproperlyConfigured(
                    "To use functional suggester backend you shall add "
                    "`FunctionalSuggesterFilterBackend` to the "
                    "`filter_backends` of your ViewSet."
                )

        queryset = self.filter_queryset(self.get_queryset())
        page = self.paginate_queryset(queryset)
        return Response(page)


class MoreLikeThisMixin(object):
    """More-like-this mixin."""

    @action(detail=True)
    def more_like_this(self, request, pk=None, id=None):
        """More-like-this functionality detail view.

        :param request:
        :return:
        """
        if "view" in request.parser_context:
            view = request.parser_context["view"]
            kwargs = copy.copy(getattr(view, "more_like_this_options", {}))
            id_ = pk if pk else id

            # Use current queryset
            queryset = self.filter_queryset(self.get_queryset())
            # We do not try to get fields from current serializer. On the
            # Elasticsearch side if no ``fields`` value is given, ``_all`` is
            # used, and although some serializers could contain less fields
            # than available, this seems like the best approach. If you want to
            # fall back to ``_all`` of Elasticsearch, leave it empty.
            fields = kwargs.pop("fields", [])
            # if not fields:
            #     serializer_class = self.get_serializer_class()
            #     fields = serializer_class.Meta.fields[:]
            if fields:
                queryset = queryset.query(
                    MoreLikeThis(
                        fields=fields,
                        like={
                            "_id": "{}".format(id_),
                            "_index": "{}".format(self.index),
                            "_type": "{}".format(self.mapping),
                        },
                        **kwargs
                    )
                ).sort("_score")
            else:
                queryset = queryset.query(
                    MoreLikeThis(
                        like={
                            "_id": "{}".format(id_),
                            "_index": "{}".format(self.index),
                            "_type": "{}".format(self.mapping),
                        },
                        **kwargs
                    )
                ).sort("_score")

            # Standard list-view implementation
            page = self.paginate_queryset(queryset)
            if page is not None:
                serializer = self.get_serializer(page, many=True)
                return self.get_paginated_response(serializer.data)

            serializer = self.get_serializer(queryset, many=True)
            return Response(serializer.data)


class BaseDocumentViewSet(ReadOnlyModelViewSet):
    """Base document ViewSet."""

    document_uid_field = "id"
    document = None  # Re-define
    pagination_class = PageNumberPagination
    dictionary_proxy = DictionaryProxy
    # permission_classes = (AllowAny,)
    ignore = []

    def __init__(self, *args, **kwargs):
        self.run_checks()

        if self.document:
            self.client = connections.get_connection(self.document._get_using())
            self.index = self.document._index._name
            self.mapping = self.document._doc_type.mapping.properties.name
            self.search = Search(
                using=self.client,
                index=self.index,
                doc_type=self.document._doc_type.name,
            )

        super(BaseDocumentViewSet, self).__init__(*args, **kwargs)

    def run_checks(self):
        assert self.document is not None

    def get_queryset(self):
        """Get queryset."""
        queryset = self.search.query()
        # Model- and object-permissions of the Django REST framework (
        # at the moment of writing they are ``DjangoModelPermissions``,
        # ``DjangoModelPermissionsOrAnonReadOnly`` and
        # ``DjangoObjectPermissions``) require ``model`` attribute to be
        # present in the queryset. Unfortunately we don't have that here.
        # The following approach seems to fix that (pretty well), since
        # model and object permissions would work out of the box (for the
        # correspondent Django model/object). Alternative ways to solve this
        # issue are: (a) set the ``_ignore_model_permissions`` to True on the
        # ``BaseDocumentViewSet`` or (b) provide alternative permission classes
        # that are almost identical to the above mentioned classes with
        # the only difference that they know how to extract the model from the
        # given queryset. If you think that chosen solution is incorrect,
        # please make an issue or submit a pull request explaining the
        # disadvantages (and ideally - propose  a better solution). Couple of
        # pros for current solution: (1) works out of the box, (2) does not
        # require modifications of current permissions (which would mean we
        # would have to keep up with permission changes of the DRF).
        queryset.model = self.document.Django.model
        return queryset

    def get_object(self):
        """Get object."""
        queryset = self.get_queryset()
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        if lookup_url_kwarg not in self.kwargs:
            raise AttributeError(
                "Expected view %s to be called with a URL keyword argument "
                "named '%s'. Fix your URL conf, or set the `.lookup_field` "
                "attribute on the view correctly."
                % (self.__class__.__name__, lookup_url_kwarg)
            )

        if lookup_url_kwarg == "id":
            get_kwargs = {"id": self.kwargs[lookup_url_kwarg]}
            if self.ignore:
                get_kwargs.update({"ignore": self.ignore})
            obj = self.document.get(**get_kwargs)

            # May raise a permission denied
            self.check_object_permissions(self.request, obj)

            if not obj and self.ignore:
                raise Http404("No result matches the given query.")
            # Remark 1: Code below works fine. Do not try to do the same as
            # done in `Remark 2`.
            dictionary_proxy = self.dictionary_proxy(
                obj.to_dict(), getattr(obj, "meta", None)
            )
            return dictionary_proxy
        else:
            queryset = queryset.filter(
                "term", **{self.document_uid_field: self.kwargs[lookup_url_kwarg]}
            )

            hits = queryset.execute().hits.hits
            count = len(hits)

            if count == 1:
                obj = hits[0]["_source"]

                # May raise a permission denied
                self.check_object_permissions(self.request, obj)

                # Remark 2: Unlike code in `Remark 1`, here we do need the
                # conditional treatment.
                if ELASTICSEARCH_GTE_7_0:
                    dictionary_proxy = self.dictionary_proxy(
                        obj.to_dict(), getattr(obj, "meta", None)
                    )
                else:
                    dictionary_proxy = self.dictionary_proxy(
                        obj, getattr(obj, "meta", None)
                    )

                return dictionary_proxy

            elif count > 1:
                raise Http404(
                    "Multiple results matches the given query. "
                    "Expected a single result."
                )

            raise Http404("No result matches the given query.")


class DocumentViewSet(BaseDocumentViewSet, SuggestMixin, FunctionalSuggestMixin):
    """DocumentViewSet with suggest and functional-suggest mix-ins."""



"""
All filter backends.
"""

from .faceted_search import FacetedSearchFilterBackend, FacetedFilterSearchFilterBackend
from .filtering import (
    FilteringFilterBackend,
    GeoSpatialFilteringFilterBackend,
    IdsFilterBackend,
    NestedFilteringFilterBackend,
    PostFilterFilteringFilterBackend,
)
from .ordering import (
    DefaultOrderingFilterBackend,
    GeoSpatialOrderingFilterBackend,
    OrderingFilterBackend,
)
from .search import (
    BaseSearchFilterBackend,
    CompoundSearchFilterBackend,
    MultiMatchSearchFilterBackend,
    SearchFilterBackend,
    SimpleQueryStringSearchFilterBackend,
)
from .source import SourceBackend
from .suggester import (
    FunctionalSuggesterFilterBackend,
    SuggesterFilterBackend,
)
from .highlight import HighlightBackend

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"



"""
Test views.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

import factories

from .base import BaseRestFrameworkTestCase

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_views"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestViews",)


@pytest.mark.django_db
class TestViews(BaseRestFrameworkTestCase):
    """Test views."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up class."""
        super(TestViews, cls).setUpClass()

        cls.books = factories.BookWithoutTagsAndOrdersFactory.create_batch(20)
        cls.tags = factories.TagGenreFactory.create_batch(20)

        cls.sleep()
        call_command("search_index", "--rebuild", "-f")

    def test_listing_view(self):
        """Test listing view."""
        url = reverse("bookdocument-list", kwargs={})
        data = {}

        # Should contain 20 results
        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), 20)

    def test_detail_view(self):
        """Test detail view."""
        __obj = self.books[0]
        url = reverse("bookdocument-detail", kwargs={"id": __obj.pk})
        data = {}

        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        for __field in ("id", "title", "pages", "state", "isbn"):
            self.assertEqual(response.data[__field], getattr(__obj, __field))

    def test_detail_view_with_custom_lookup(self):
        """Test detail view with a custom lookup field."""
        __obj = self.tags[0]
        url = reverse("tagdocument-detail", kwargs={"title": __obj.title})
        data = {}

        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["id"], __obj.title)


if __name__ == "__main__":
    unittest.main()



"""
Highlight backend.
"""

from rest_framework.filters import BaseFilterBackend

__title__ = "django_elasticsearch_dsl_drf_alt.highlight"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("HighlightBackend",)


class HighlightBackend(BaseFilterBackend):
    """Highlight backend.

    Example:

        >>> from django_elasticsearch_dsl_drf_alt.filter_backends import (
        >>>     HighlightBackend
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.viewsets import (
        >>>     BaseDocumentViewSet,
        >>> )
        >>>
        >>> # Local article document definition
        >>> from .documents import ArticleDocument
        >>>
        >>> # Local article document serializer
        >>> from .serializers import ArticleDocumentSerializer
        >>>
        >>> class ArticleDocumentView(BaseDocumentViewSet):
        >>>
        >>>     document = ArticleDocument
        >>>     serializer_class = ArticleDocumentSerializer
        >>>     filter_backends = [HighlightBackend,]
        >>>     highlight_fields = {
        >>>         'author.name': {
        >>>             'enabled': False,
        >>>             'options': {
        >>>                 'fragment_size': 150,
        >>>                 'number_of_fragments': 3
        >>>             }
        >>>         }
        >>>         'title': {
        >>>             'options': {
        >>>                 'pre_tags' : ["<em>"],
        >>>                 'post_tags' : ["</em>"]
        >>>             },
        >>>             'enabled': True,
        >>>         },
        >>>     }

    Highlight make queries to be more heavy. That's why by default all
    highlights are disabled and enabled only explicitly either in the filter
    options (`enabled` set to True) or via query params
    `?highlight=author.name&highlight=title`.
    """

    highlight_param = "highlight"

    @classmethod
    def prepare_highlight_fields(cls, view):
        """Prepare faceted search fields.

        Prepares the following structure:

            >>> {
            >>>     'author.name': {
            >>>         'enabled': False,
            >>>         'options': {
            >>>             'fragment_size': 150,
            >>>             'number_of_fragments': 3
            >>>         }
            >>>     }
            >>>     'title': {
            >>>         'options': {
            >>>             'pre_tags' : ["<em>"],
            >>>             'post_tags' : ["</em>"]
            >>>         },
            >>>         'enabled': True,
            >>>     },
            >>> }

        :param view:
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Highlight fields options.
        :rtype: dict
        """
        highlight_fields = view.highlight_fields

        for field, options in highlight_fields.items():
            if "enabled" not in highlight_fields[field]:
                highlight_fields[field]["enabled"] = False

            if "options" not in highlight_fields[field]:
                highlight_fields[field]["options"] = {}

        return highlight_fields

    def get_highlight_query_params(self, request):
        """Get highlight query params.

        :param request: Django REST framework request.
        :type request: rest_framework.request.Request
        :return: List of search query params.
        :rtype: list
        """
        query_params = request.query_params.copy()
        return query_params.getlist(self.highlight_param, [])

    def filter_queryset(self, request, queryset, view):
        """Filter the queryset.

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        highlight_query_params = self.get_highlight_query_params(request)
        highlight_fields = self.prepare_highlight_fields(view)
        for __field, __options in highlight_fields.items():
            if __field in highlight_query_params or __options["enabled"]:
                queryset = queryset.highlight(__field, **__options["options"])

        return queryset



"""
Nested filtering backend.
"""

from elasticsearch_dsl.query import Q
from django.core.exceptions import ImproperlyConfigured
from django_elasticsearch_dsl import fields

from six import string_types

from ...constants import (
    ALL_LOOKUP_FILTERS_AND_QUERIES,
    LOOKUP_FILTER_TERMS,
)

from ...compat import coreapi
from ...compat import coreschema
from .common import FilteringFilterBackend

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.filtering.nested"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("NestedFilteringFilterBackend",)


class NestedFilteringFilterBackend(FilteringFilterBackend):
    """Nested filter backend.

    Example:

        >>> from django_elasticsearch_dsl_drf_alt.constants import (
        >>>     LOOKUP_FILTER_TERM,
        >>>     LOOKUP_FILTER_PREFIX,
        >>>     LOOKUP_FILTER_WILDCARD,
        >>>     LOOKUP_QUERY_EXCLUDE,
        >>>     LOOKUP_QUERY_ISNULL,
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.filter_backends import (
        >>>     NestedFilteringFilterBackend
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.viewsets import (
        >>>     BaseDocumentViewSet,
        >>> )
        >>>
        >>> # Local article document definition
        >>> from .documents import ArticleDocument
        >>>
        >>> # Local article document serializer
        >>> from .serializers import ArticleDocumentSerializer
        >>>
        >>> class ArticleDocumentView(BaseDocumentViewSet):
        >>>
        >>>     document = ArticleDocument
        >>>     serializer_class = ArticleDocumentSerializer
        >>>     filter_backends = [NestedFilteringFilterBackend,]
        >>>     nested_filter_fields = {
        >>>         'country': {
        >>>             'field': 'continent.country.name.raw',
        >>>             'path': 'continent.country',
        >>>             'lookups': [
        >>>                 LOOKUP_FILTER_TERM,
        >>>                 LOOKUP_FILTER_TERMS,
        >>>                 LOOKUP_FILTER_PREFIX,
        >>>                 LOOKUP_FILTER_WILDCARD,
        >>>                 LOOKUP_QUERY_EXCLUDE,
        >>>                 LOOKUP_QUERY_ISNULL,
        >>>             ],
        >>>         }
        >>> }
    """

    @classmethod
    def prepare_filter_fields(cls, view):
        """Prepare filter fields.

        :param view:
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Filtering options.
        :rtype: dict
        """
        if not hasattr(view, "nested_filter_fields"):
            raise ImproperlyConfigured(
                "You need to define `nested_filter_fields` in your `{}` view "
                "when using `{}` filter backend."
                "".format(view.__class__.__name__, cls.__name__)
            )

        filter_fields = view.nested_filter_fields

        for field, options in filter_fields.items():
            if options is None or isinstance(options, string_types):
                filter_fields[field] = {"field": options or field}
            elif "field" not in filter_fields[field]:
                filter_fields[field]["field"] = field

            if "lookups" not in filter_fields[field]:
                filter_fields[field]["lookups"] = tuple(ALL_LOOKUP_FILTERS_AND_QUERIES)

        return filter_fields

    def get_filter_field_nested_path(self, filter_fields, field_name):
        """Get filter field path to be used in nested query.

        :param filter_fields:
        :param field_name:
        :return:
        """
        if "path" in filter_fields[field_name]:
            return filter_fields[field_name]["path"]
        return field_name

    def get_filter_query_params(self, request, view):
        """Get query params to be filtered on.

        :param request: Django REST framework request.
        :param view: View.
        :type request: rest_framework.request.Request
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Request query params to filter on.
        :rtype: dict
        """
        query_params = request.query_params.copy()

        filter_query_params = {}
        filter_fields = self.prepare_filter_fields(view)
        for query_param in query_params:
            query_param_list = self.split_lookup_filter(query_param, maxsplit=1)
            field_name = query_param_list[0]

            if field_name in filter_fields:
                lookup_param = None
                if len(query_param_list) > 1:
                    lookup_param = query_param_list[1]

                valid_lookups = filter_fields[field_name]["lookups"]
                nested_path = self.get_filter_field_nested_path(
                    filter_fields, field_name
                )

                if lookup_param is None or lookup_param in valid_lookups:
                    values = [
                        __value.strip()
                        for __value in query_params.getlist(query_param)
                        if __value.strip() != ""
                    ]

                    if values:
                        filter_query_params[query_param] = {
                            "lookup": lookup_param,
                            "values": values,
                            "field": filter_fields[field_name].get("field", field_name),
                            "type": view.mapping,
                            "path": nested_path,
                        }
        return filter_query_params

    @classmethod
    def apply_filter(cls, queryset, options=None, args=None, kwargs=None):
        """Apply filter.

        :param queryset:
        :param options:
        :param args:
        :param kwargs:
        :return:
        """
        if options is None or "path" not in options:
            raise ImproperlyConfigured(
                "You should provide an `path` argument in the field options."
            )

        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        return queryset.query(
            "nested", path=options.get("path"), query=Q(*args, **kwargs)
        )

    @classmethod
    def apply_query(cls, queryset, options=None, args=None, kwargs=None):
        """Apply query.

        :param queryset:
        :param options:
        :param args:
        :param kwargs:
        :return:
        """
        if options is None:
            raise ImproperlyConfigured(
                "You should provide an `path` argument in the field options."
            )

        path = options.pop("path")

        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        return queryset.query("nested", path=path, query=Q(*args, **kwargs))

    def get_coreschema_field(self, field):
        if isinstance(field, fields.IntegerField):
            field_cls = coreschema.Number
        else:
            field_cls = coreschema.String
        return field_cls()

    def get_schema_fields(self, view):
        assert coreapi is not None, (
            "coreapi must be installed to " "use `get_schema_fields()`"
        )
        assert coreschema is not None, (
            "coreschema must be installed to " "use `get_schema_fields()`"
        )
        filter_fields = getattr(view, "nested_filter_fields", None)
        document = getattr(view, "document", None)

        return (
            []
            if not filter_fields
            else [
                coreapi.Field(
                    name=field_name,
                    required=False,
                    location="query",
                    schema=self.get_coreschema_field(document._fields.get(field_name)),
                )
                for field_name in filter_fields
            ]
        )



"""
Geo-spatial ordering backend.
"""

from rest_framework.filters import BaseFilterBackend

from ..mixins import FilterBackendMixin
from ...constants import (
    GEO_DISTANCE_ORDERING_PARAM,
)

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.ordering.common"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("GeoSpatialOrderingFilterBackend",)


class GeoSpatialOrderingFilterBackend(BaseFilterBackend, FilterBackendMixin):
    """Geo-spatial ordering filter backend for Elasticsearch.

    Example:

        >>> from django_elasticsearch_dsl_drf_alt.filter_backends import (
        >>>     GeoSpatialOrderingFilterBackend
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.viewsets import (
        >>>     BaseDocumentViewSet,
        >>> )
        >>>
        >>> # Local article document definition
        >>> from .documents import ArticleDocument
        >>>
        >>> # Local article document serializer
        >>> from .serializers import ArticleDocumentSerializer
        >>>
        >>> class ArticleDocumentView(BaseDocumentViewSet):
        >>>
        >>>     document = ArticleDocument
        >>>     serializer_class = ArticleDocumentSerializer
        >>>     filter_backends = [GeoSpatialOrderingFilterBackend,]
        >>>     geo_spatial_ordering_fields = {
        >>>         'location': {
        >>>             'field': 'location',
        >>>         }
        >>>     }
    """

    ordering_param = GEO_DISTANCE_ORDERING_PARAM

    @classmethod
    def get_geo_distance_params(cls, value, field):
        """Get params for `geo_distance` ordering.

        Example:

            /api/articles/?ordering=-location__45.3214__-34.3421__km__planes

        :param value:
        :param field:
        :type value: str
        :type field:
        :return: Params to be used in `geo_distance` query.
        :rtype: dict
        """
        __values = cls.split_lookup_complex_value(value, maxsplit=3)
        __len_values = len(__values)

        if __len_values < 2:
            return {}

        params = {
            "_geo_distance": {
                field: {
                    "lat": __values[0],
                    "lon": __values[1],
                }
            }
        }

        if __len_values > 2:
            params["_geo_distance"]["unit"] = __values[2]
        else:
            params["_geo_distance"]["unit"] = "m"
        if __len_values > 3:
            params["_geo_distance"]["distance_type"] = __values[3]
        else:
            params["_geo_distance"]["distance_type"] = "arc"

        return params

    def get_geo_spatial_field_name(self, request, view, name):
        """Get geo-spatial field name.

        We have to deal with a couple of situations here:

        Example 1:

         >>> geo_spatial_ordering_fields = {
         >>>     'location': None,
         >>> }

        Example 2:

        >>> geo_spatial_ordering_fields = {
        >>>     'location': 'location',
        >>> }

        Example 3:

        >>> geo_spatial_ordering_fields = {
        >>>     'location': {
        >>>         'field': 'location'
        >>>     },
        >>> }

        :param request:
        :param view:
        :param name:
        :return:
        """
        options = view.geo_spatial_ordering_fields[name]
        if options is None:
            return name
        elif isinstance(options, dict):
            if "field" in options:
                return options["field"]
        else:
            return options

    def get_ordering_query_params(self, request, view):
        """Get ordering query params.

        :param request: Django REST framework request.
        :param view: View.
        :type request: rest_framework.request.Request
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Ordering params to be used for ordering.
        :rtype: list
        """
        # TODO: Support `mode` argument.
        query_params = request.query_params.copy()
        ordering_query_params = query_params.getlist(self.ordering_param, [])
        __ordering_params = []
        # Remove invalid ordering query params
        for query_param in ordering_query_params:
            try:
                __key, __value = FilterBackendMixin.split_lookup_complex_value(
                    query_param.lstrip("-"),
                    maxsplit=1,
                )
            # Probably using both
            # OrderingFilterBackend/DefaultOrderingFilterBackend
            # and GeoSpatialOrderingFilterBackend
            except ValueError:
                break
            __direction = "desc" if query_param.startswith("-") else "asc"
            if __key in view.geo_spatial_ordering_fields:
                __field_name = self.get_geo_spatial_field_name(request, view, __key)
                __params = self.get_geo_distance_params(__value, __field_name)
                __params["_geo_distance"]["order"] = __direction
                __ordering_params.append(__params)

        return __ordering_params

    def filter_queryset(self, request, queryset, view):
        """Filter the queryset.

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        ordering_query_params = self.get_ordering_query_params(request, view)

        if ordering_query_params:
            return queryset.sort(*ordering_query_params)

        return queryset



"""
Term level filtering and ``post_filter`` backends.
"""

from .common import FilteringFilterBackend
from .geo_spatial import GeoSpatialFilteringFilterBackend
from .ids import IdsFilterBackend
from .nested import NestedFilteringFilterBackend
from .post_filter import PostFilterFilteringFilterBackend

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.filtering"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "FilteringFilterBackend",
    "GeoSpatialFilteringFilterBackend",
    "IdsFilterBackend",
    "NestedFilteringFilterBackend",
    "PostFilterFilteringFilterBackend",
)



# """
# Aggregations filtering backends.
# """
#
# from .bucket_aggregations import *
# from .metrics_aggregations import *
# from .pipeline_aggregations import *
#
# __title__ = 'django_elasticsearch_dsl_drf_alt.filter_backends.suggester'
# __author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
# __copyright__ = '2017-2020 Artur Barseghyan'
# __license__ = 'GPL 2.0/LGPL 2.1'
# __all__ = (
#     'BucketAggregationsFilterBackend',
#     'MetricsAggregationsFilterBackend',
#     'PipelineAggregationsFilterBackend',
# )



"""
Faceted search backend.
"""

import copy
from collections import defaultdict

from elasticsearch_dsl import TermsFacet
from elasticsearch_dsl.query import Q

from rest_framework.filters import BaseFilterBackend

from six import string_types, iteritems

__title__ = "django_elasticsearch_dsl_drf_alt.faceted_search"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("FacetedSearchFilterBackend", "FacetedFilterSearchFilterBackend")

from django_elasticsearch_dsl_drf_alt.filter_backends.filtering import (
    FilteringFilterBackend,
)


class FacetedSearchFilterBackend(BaseFilterBackend):
    """Faceted search backend.

    Example:

        >>> from django_elasticsearch_dsl_drf_alt.filter_backends import (
        >>>     FacetedSearchFilterBackend
        >>> )
        >>> from elasticsearch_dsl import TermsFacet, DateHistogramFacet
        >>> from django_elasticsearch_dsl_drf_alt.viewsets import (
        >>>     BaseDocumentViewSet,
        >>> )
        >>>
        >>> # Local article document definition
        >>> from .documents import ArticleDocument
        >>>
        >>> # Local article document serializer
        >>> from .serializers import ArticleDocumentSerializer
        >>>
        >>> class ArticleDocumentView(BaseDocumentViewSet):
        >>>
        >>>     document = ArticleDocument
        >>>     serializer_class = ArticleDocumentSerializer
        >>>     filter_backends = [FacetedSearchFilterBackend,]
        >>>     faceted_search_fields = {
        >>>         'title': 'title.raw',  # Uses `TermsFacet` by default
        >>>         'state': {
        >>>             'field': 'state.raw',
        >>>             'facet': TermsFacet,
        >>>         },
        >>>         'publisher': {
        >>>             'field': 'publisher.raw',
        >>>             'facet': TermsFacet,
        >>>             'enabled': False,
        >>>         },
        >>>         'date_published': {
        >>>             'field': 'date_published.raw',
        >>>             'facet': DateHistogramFacet,
        >>>             'options': {
        >>>                 'interval': 'month',
        >>>             },
        >>>             'enabled': True,
        >>>         },
        >>>
        >>>     }

    Facets make queries to be more heavy. That's why by default all
    facets are disabled and enabled only explicitly either in the filter
    options (`enabled` set to True) or via query params
    `?facet=state&facet=date_published`.
    """

    faceted_search_param = "facet"

    @classmethod
    def prepare_faceted_search_fields(cls, view):
        """Prepare faceted search fields.

        Prepares the following structure:

            >>> {
            >>>     'publisher': {
            >>>         'field': 'publisher.raw',
            >>>         'facet': TermsFacet,
            >>>         'enabled': False,
            >>>     }
            >>>     'date_published': {
            >>>         'field': 'date_published.raw',
            >>>         'facet': DateHistogramFacet,
            >>>         'options': {
            >>>             'interval': 'month',
            >>>         },
            >>>         'enabled': True,
            >>>     },
            >>> }

        :param view:
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Faceted search fields options.
        :rtype: dict
        """
        faceted_search_fields = copy.deepcopy(view.faceted_search_fields)

        for field, options in faceted_search_fields.items():
            if options is None or isinstance(options, string_types):
                faceted_search_fields[field] = {"field": options or field}
            elif "field" not in faceted_search_fields[field]:
                faceted_search_fields[field]["field"] = field

            if "enabled" not in faceted_search_fields[field]:
                faceted_search_fields[field]["enabled"] = False

            if "facet" not in faceted_search_fields[field]:
                faceted_search_fields[field]["facet"] = TermsFacet

            if "options" not in faceted_search_fields[field]:
                faceted_search_fields[field]["options"] = {}

            faceted_search_fields[field]["global"] = faceted_search_fields[field].get(
                "global", False
            )

        return faceted_search_fields

    def get_faceted_search_query_params(self, request):
        """Get faceted search query params.

        :param request: Django REST framework request.
        :type request: rest_framework.request.Request
        :return: List of search query params.
        :rtype: list
        """
        query_params = request.query_params.copy()
        return query_params.getlist(self.faceted_search_param, [])

    def construct_facets(self, request, view):
        """Construct facets.

        Turns the following structure:

            >>> {
            >>>     'publisher': {
            >>>         'field': 'publisher.raw',
            >>>         'facet': TermsFacet,
            >>>         'enabled': False,
            >>>     }
            >>>     'date_published': {
            >>>         'field': 'date_published',
            >>>         'facet': DateHistogramFacet,
            >>>         'options': {
            >>>             'interval': 'month',
            >>>         },
            >>>         'enabled': True,
            >>>     },
            >>> }

        Into the following structure:

            >>> {
            >>>     'publisher': TermsFacet(field='publisher.raw'),
            >>>     'publishing_frequency': DateHistogramFacet(
            >>>         field='date_published.raw',
            >>>         interval='month'
            >>>     ),
            >>> }
        """
        __facets = {}
        faceted_search_query_params = self.get_faceted_search_query_params(request)
        faceted_search_fields = self.prepare_faceted_search_fields(view)
        for __field, __options in faceted_search_fields.items():
            if __field in faceted_search_query_params or __options["enabled"]:
                __facets.update(
                    {
                        __field: {
                            "facet": faceted_search_fields[__field]["facet"](
                                field=faceted_search_fields[__field]["field"],
                                **faceted_search_fields[__field]["options"]
                            ),
                            "global": faceted_search_fields[__field]["global"],
                        }
                    }
                )
        return __facets

    def aggregate(self, request, queryset, view):
        """Aggregate.

        :param request:
        :param queryset:
        :param view:
        :return:
        """
        __facets = self.construct_facets(request, view)
        for __field, __facet in iteritems(__facets):
            agg = __facet["facet"].get_aggregation()
            agg_filter = Q("match_all")

            # TODO: Implement
            # for __filter_field, __filter in iteritems(self._filters):
            #     if __field == __filter_field:
            #         continue
            #     agg_filter &= __filter

            if __facet["global"]:
                queryset.aggs.bucket("_filter_" + __field, "global").bucket(
                    __field, agg
                )
            else:
                queryset.aggs.bucket(
                    "_filter_" + __field, "filter", filter=agg_filter
                ).bucket(__field, agg)

        return queryset

    def filter_queryset(self, request, queryset, view):
        """Filter the queryset.

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return self.aggregate(request, queryset, view)


class FacetedFilterSearchFilterBackend(
    FilteringFilterBackend, FacetedSearchFilterBackend
):
    """Combined faceting and filtering backend similar to elasticsearch-dsl's FacetedSearch class.
    It combines the functionality of FilteringFilterBackend and FacetedSearchFilterBackend to take filters into
    account when creating facets.

    This backend uses the same configuration fields as FilteringFilterBackend and FacetedSearchFilterBackend.
    This backend replaces their functionality and should not be used together with either of those backends.

    Note that to work correctly, the actual elasticsearch field must be the same for a facet and its matching filter.
    For example, if a facet will aggregate on field `state.raw`, then the filter must also filter on `state.raw`,
    and not just `state`.

    When creating a facet, filters for faceted fields other than for the current facet are applied. Filters
    for faceted fields are then applied as post_filters. Filters on non-faceted fields are applied as normal filters.
    """

    def filter_queryset(self, request, queryset, view):
        # the fact that apply_filter is a classmethod means we can't store state on self,
        # so we hitch it onto queryset
        queryset._facets = self.construct_facets(request, view)
        queryset._faceted_fields = set(
            f["facet"]._params["field"] for f in queryset._facets.values()
        )
        queryset._filters = defaultdict(list)

        # apply filters
        queryset = FilteringFilterBackend.filter_queryset(self, request, queryset, view)

        # apply aggregations
        return self.aggregate(request, queryset, view)

    @classmethod
    def apply_filter(cls, queryset, options=None, args=None, kwargs=None):
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        facets = queryset._facets
        faceted_fields = queryset._faceted_fields
        filters = queryset._filters

        # if this field is faceted, then apply it as a post-filter
        if options["field"] in faceted_fields:
            queryset = queryset.post_filter(*args, **kwargs)
        else:
            queryset = queryset.filter(*args, **kwargs)

        filters[options["field"]].append(Q(*args, **kwargs))

        # ensure the new queryset object retains the helper variables
        queryset._facets = facets
        queryset._faceted_fields = faceted_fields
        queryset._filters = filters
        return queryset

    def aggregate(self, request, queryset, view):
        facets = queryset._facets
        faceted_fields = queryset._faceted_fields
        filters = queryset._filters

        for field, facet in facets.items():
            agg = facet["facet"].get_aggregation()

            if facet["global"]:
                queryset.aggs.bucket("_filter_" + field, "global").bucket(field, agg)
                continue

            agg_filter = Q("match_all")
            for f, _filter in filters.items():
                # apply filters for that are applicable for facets other than this one
                if agg.field == f or f not in faceted_fields:
                    continue
                # combine with or
                q = _filter[0]
                for x in _filter[1:]:
                    q = q | x
                agg_filter &= q

            queryset.aggs.bucket(
                "_filter_" + field, "filter", filter=agg_filter
            ).bucket(field, agg)

        return queryset



"""
Helpers.
"""

from elasticsearch_dsl.utils import AttrDict, AttrList

__title__ = "django_elasticsearch_dsl_drf_alt.fields.helpers"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("to_representation",)


def to_representation(value):
    """To representation."""
    if isinstance(value, AttrDict):
        return value.to_dict()
    if isinstance(value, AttrList):
        _value = [to_representation(__v) for __v in value]
        return _value
        # If approach above doesn't work, replace it with the code below
        # try:
        #     _value = list(value)
        #     json.dumps(_value)
        #     return _value
        # except TypeError:
        #     _value = [to_representation(__v) for __v in value]
        #     return _value
    return value



"""
Test nested filtering backend.
"""

from __future__ import absolute_import

import copy
import unittest

from django.core.management import call_command

import pytest

from rest_framework import status

from search_indexes.viewsets import AddressDocumentViewSet

from ..constants import SEPARATOR_LOOKUP_COMPLEX_VALUE
from ..filter_backends import NestedFilteringFilterBackend
from .base import (
    BaseRestFrameworkTestCase,
    CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED,
    CORE_API_AND_CORE_SCHEMA_MISSING_MSG,
)
from .data_mixins import AddressesMixin

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_filtering_nested"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestFilteringNested",)


@pytest.mark.django_db
class TestFilteringNested(BaseRestFrameworkTestCase, AddressesMixin):
    """Test filtering nested."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up."""
        super(TestFilteringNested, cls).setUpClass()
        # Testing nested objects: Addresses, cities and countries, continents
        cls.created_addresses()

        cls.all_addresses = (
            cls.addresses_in_yerevan
            + cls.addresses_in_amsterdam
            + cls.addresses_in_dublin
            + cls.addresses_in_yeovil
            + cls.addresses_in_buenos_aires
        )
        cls.all_addresses_ids = sorted(set([obj.id for obj in cls.all_addresses]))
        cls.all_cities_ids = sorted(set([obj.city.id for obj in cls.all_addresses]))
        cls.add_addresses_dict = {
            cls.addresses_in_yerevan[0].city.id: {
                "count": len(cls.addresses_in_yerevan),
                "objects": cls.addresses_in_yerevan,
            },
            cls.addresses_in_amsterdam[0].city.id: {
                "count": len(cls.addresses_in_amsterdam),
                "objects": cls.addresses_in_amsterdam,
            },
            cls.addresses_in_dublin[0].city.id: {
                "count": len(cls.addresses_in_dublin),
                "objects": cls.addresses_in_dublin,
            },
            cls.addresses_in_yeovil[0].city.id: {
                "count": len(cls.addresses_in_yeovil),
                "objects": cls.addresses_in_yeovil,
            },
            cls.addresses_in_buenos_aires[0].city.id: {
                "count": len(cls.addresses_in_buenos_aires),
                "objects": cls.addresses_in_buenos_aires,
            },
        }

        cls.sleep()
        # Update the Elasticsearch index
        call_command("search_index", "--rebuild", "-f")

        # Testing coreapi and coreschema
        cls.backend = NestedFilteringFilterBackend()
        cls.view = AddressDocumentViewSet()

    @property
    def base_url(self):
        return self.addresses_url

    # ***********************************************************************
    # ************************ Simple fields ********************************
    # ***********************************************************************

    def _field_filter_value(self, field_name, value, count):
        """Field filter value.

        Usage example:

            >>> self._field_filter_value(
            >>>     'continent_country_city__contains',
            >>>     'ere',
            >>>     self.addresses_in_yerevan_count
            >>> )
        """
        url = self.base_url[:]
        data = {}
        response = self.client.get(url + "?{}={}".format(field_name, value), data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), count)

    def _field_filter_term(
        self, field_name, filter_value, total_num_results, filtered_num_results
    ):
        """Field filter term.

        Example:

            http://localhost:8000/api/articles/?tags=children
        """
        self.authenticate()

        url = self.base_url[:]
        data = {}

        # Should contain `total_num_results` results
        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), total_num_results)

        # Should contain only `filtered_num_results` results
        filtered_response = self.client.get(
            url + "?{}={}".format(field_name, filter_value), data
        )
        self.assertEqual(filtered_response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(filtered_response.data["results"]), filtered_num_results)

    def test_field_filter_term(self):
        """Field filter term."""
        return self._field_filter_term(
            "continent_country_city",
            "Yerevan",
            self.all_addresses_count,
            self.addresses_in_yerevan_count,
        )

    def test_field_filter_term_explicit(self):
        """Field filter term."""
        return self._field_filter_term(
            "continent_country_city__term",
            "Yerevan",
            self.all_addresses_count,
            self.addresses_in_yerevan_count,
        )

    def test_field_filter_range(self):
        """Field filter range.

        Example:

            http://localhost:8000/api/users/?age__range=16__67
        """
        # Pick the first and the last elements from the list
        lower_id = self.all_cities_ids[1]
        upper_id = self.all_cities_ids[3]
        value = "{lower_id}{sep}{upper_id}".format(
            lower_id=lower_id, upper_id=upper_id, sep=SEPARATOR_LOOKUP_COMPLEX_VALUE
        )
        # Calculate expected number of items
        count = (
            self.add_addresses_dict[lower_id]["count"]
            + self.add_addresses_dict[lower_id + 1]["count"]
            + self.add_addresses_dict[upper_id]["count"]
        )
        return self._field_filter_value(
            "continent_country_city_id__range", value, count
        )

    def test_field_filter_range_with_boost(self):
        """Field filter range.

        Example:

            http://localhost:8000/api/users/?age__range=16__67__2.0
        """
        lower_id = self.all_cities_ids[2]
        upper_id = self.all_cities_ids[4]
        value = "{lower_id}{sep}{upper_id}{sep}{boost}".format(
            lower_id=lower_id,
            upper_id=upper_id,
            boost="2.0",
            sep=SEPARATOR_LOOKUP_COMPLEX_VALUE,
        )
        # Calculate expected number of items
        count = (
            self.add_addresses_dict[lower_id]["count"]
            + self.add_addresses_dict[lower_id + 1]["count"]
            + self.add_addresses_dict[upper_id]["count"]
        )
        return self._field_filter_value(
            "continent_country_city_id__range", value, count
        )

    def test_field_filter_prefix(self):
        """Test filter prefix.

        Example:

            http://localhost:8000/api/articles/?tags__prefix=bio
        """
        return self._field_filter_value(
            "continent_country_city__prefix",
            "Ye",  # Matches Yerevan and Yeovil
            self.addresses_in_yeovil_count + self.addresses_in_yerevan_count,
        )

    def test_field_filter_in(self):
        """Test filter in.

        Example:

            http://localhost:8000/api/articles/?id__in=1__2__3
        """
        ids = [
            self.addresses_in_yerevan[0].city.id,
            self.addresses_in_amsterdam[0].city.id,
        ]
        return self._field_filter_value(
            "continent_country_city_id__in",
            SEPARATOR_LOOKUP_COMPLEX_VALUE.join([str(_id) for _id in ids]),
            self.addresses_in_amsterdam_count + self.addresses_in_yerevan_count,
        )

    def _field_filter_terms_list(self, field_name, in_values, count):
        """Field filter terms.

        Example:

            http://localhost:8000/api/articles/?id=1&id=2&id=3
        """
        url = self.base_url[:]
        data = {}
        url_parts = ["{}={}".format(field_name, val) for val in in_values]
        response = self.client.get(url + "?{}".format("&".join(url_parts)), data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), count)

    def test_field_filter_terms_list(self):
        """Test filter terms."""
        return self._field_filter_terms_list(
            "continent_country_city",
            ["Yerevan", "Amsterdam"],
            self.addresses_in_amsterdam_count + self.addresses_in_yerevan_count,
        )

    def test_field_filter_terms_string(self):
        """Test filter terms.

        Example:

            http://localhost:8000/api/articles/?id__terms=1__2__3
        """
        return self._field_filter_value(
            "continent_country_city__terms",
            SEPARATOR_LOOKUP_COMPLEX_VALUE.join(["Yerevan", "Dublin"]),
            self.addresses_in_dublin_count + self.addresses_in_yerevan_count,
        )

    def test_field_filter_exists_true(self):
        """Test filter exists true.

        Example:

            http://localhost:8000/api/articles/?tags__exists=true
        """
        return self._field_filter_value(
            "continent_country_city__exists", "true", self.all_addresses_count
        )

    def test_field_filter_exists_false(self):
        """Test filter exists.

        Example:

            http://localhost:8000/api/articles/?non_existent__exists=false
        """
        return self._field_filter_value(
            "non_existent_field__exists", "false", self.all_addresses_count
        )

    def test_field_filter_wildcard(self):
        """Test filter wildcard.

        Example:

            http://localhost:8000/api/articles/?title__wildcard=*elusional*
        """
        self._field_filter_value(
            "continent_country_city__wildcard",
            "*{}*".format("ere"),  # Matches Yerevan
            self.addresses_in_yerevan_count,
        )

        return self._field_filter_value(
            "continent_country_city__wildcard",
            "*{}*".format("ire"),  # Matches Buenos Aires
            self.addresses_in_buenos_aires_count,
        )

    def test_field_filter_exclude(self):
        """Test filter exclude.

        Example:

            http://localhost:8000/api/articles/?tags__exclude=children
        """
        return self._field_filter_value(
            "continent_country_city__exclude",
            "Yeovil",
            self.all_addresses_count - self.addresses_in_yeovil_count,
        )

    # def test_field_filter_isnull_true(self):
    #     """Test filter isnull true.
    #
    #     Example:
    #
    #         http://localhost:8000/api/articles/?null_field__isnull=true
    #     """
    #     self._field_filter_value(
    #         'null_field__isnull',
    #         'true',
    #         self.all_count
    #     )
    #     self._field_filter_value(
    #         'tags__isnull',
    #         'true',
    #         self.no_tags_count
    #     )
    #
    # def test_field_filter_isnull_false(self):
    #     """Test filter isnull true.
    #
    #     Example:
    #
    #         http://localhost:8000/api/articles/?tags__isnull=false
    #     """
    #     self._field_filter_value(
    #         'null_field__isnull',
    #         'false',
    #         0
    #     )
    #     self._field_filter_value(
    #         'tags__isnull',
    #         'false',
    #         self.all_count - self.no_tags_count
    #     )

    def test_field_filter_endswith(self):
        """Test filter endswith.

        Example:

            http://localhost:8000/api/articles/?state__endswith=lished
        """
        self._field_filter_value(
            "continent_country_city__endswith",
            "dam",  # Matches Amsterdam
            self.addresses_in_amsterdam_count,
        )
        return self._field_filter_value(
            "continent_country_city__endswith",
            "van",  # Matches Yerevan
            self.addresses_in_yerevan_count,
        )

    def test_field_filter_contains(self):
        """Test filter contains.

        Example:

            http://localhost:8000/api/articles/?state__contains=lishe
        """
        self._field_filter_value(
            "continent_country_city__contains",
            "ster",  # Matches Amsterdam
            self.addresses_in_amsterdam_count,
        )

        return self._field_filter_value(
            "continent_country_city__contains",
            "bli",  # Matches Dublin
            self.addresses_in_dublin_count,
        )

    # def _field_filter_gte_lte(self, field_name, value, lookup, boost=None):
    #     """Field filter gt/gte/lt/lte.
    #
    #     Example:
    #
    #         http://localhost:8000/api/users/?id__gt=10
    #         http://localhost:8000/api/users/?id__gte=10
    #         http://localhost:8000/api/users/?id__lt=10
    #         http://localhost:8000/api/users/?id__lte=10
    #     """
    #     url = self.base_url[:]
    #     data = {}
    #
    #     if boost is not None:
    #         url += '?{}__{}={}|{}'.format(field_name, lookup, value, boost)
    #     else:
    #         url += '?{}__{}={}'.format(field_name, lookup, value)
    #
    #     response = self.client.get(
    #         url,
    #         data
    #     )
    #
    #     self.assertEqual(response.status_code, status.HTTP_200_OK)
    #
    #     __mapping = {
    #         'gt': self.assertGreater,
    #         'gte': self.assertGreaterEqual,
    #         'lt': self.assertLess,
    #         'lte': self.assertLessEqual,
    #     }
    #     __func = __mapping.get(lookup)
    #
    #     if callable(__func):
    #         for obj in response.data['results']:
    #             __func(
    #                 obj['id'],
    #                 value
    #             )
    #
    # def test_field_filter_gt(self):
    #     """Field filter gt.
    #
    #     Example:
    #
    #         http://localhost:8000/api/users/?id__gt=10
    #     :return:
    #     """
    #     return self._field_filter_gte_lte('id', self.in_progress[0].id, 'gt')
    #
    # def test_field_filter_gt_with_boost(self):
    #     """Field filter gt with boost.
    #
    #     Example:
    #
    #         http://localhost:8000/api/users/?id__gt=10|2.0
    #     :return:
    #     """
    #     # TODO: check boost value
    #     return self._field_filter_gte_lte(
    #         'id',
    #         self.in_progress[0].id,
    #         'gt',
    #         '2.0'
    #     )
    #
    # def test_field_filter_gte(self):
    #     """Field filter gte.
    #
    #     Example:
    #
    #         http://localhost:8000/api/users/?id__gte=10
    #     :return:
    #     """
    #     return self._field_filter_gte_lte(
    #         'id', self.in_progress[0].id, 'gte'
    #     )
    #
    # def test_field_filter_lt(self):
    #     """Field filter lt.
    #
    #     Example:
    #
    #         http://localhost:8000/api/users/?id__lt=10
    #     :return:
    #     """
    #     return self._field_filter_gte_lte('id', self.in_progress[0].id, 'lt')
    #
    # def test_field_filter_lt_with_boost(self):
    #     """Field filter lt with boost.
    #
    #     Example:
    #
    #         http://localhost:8000/api/users/?id__lt=10|2.0
    #     :return:
    #     """
    #     # TODO: check boost value
    #     return self._field_filter_gte_lte(
    #         'id',
    #         self.in_progress[0].id,
    #         'lt',
    #         '2.0'
    #     )
    #
    # def test_field_filter_lte(self):
    #     """Field filter lte.
    #
    #     Example:
    #
    #         http://localhost:8000/api/users/?id__lte=10
    #     :return:
    #     """
    #     return self._field_filter_gte_lte(
    #         'id', self.in_progress[0].id, 'lte'
    #     )
    #
    # def test_ids_filter(self):
    #     """Test ids filter.
    #
    #     Example:
    #
    #         http://localhost:8000/api/articles/?ids=68|64|58
    #         http://localhost:8000/api/articles/?ids=68&ids=64&ids=58
    #     """
    #     __ids = [str(__obj.id) for __obj in self.published]
    #     return self._field_filter_value(
    #         'ids',
    #         '|'.join(__ids),
    #         self.published_count
    #     )

    # ***********************************************************************
    # ******************** Core api and core schema *************************
    # ***********************************************************************

    @unittest.skipIf(
        not CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED, CORE_API_AND_CORE_SCHEMA_MISSING_MSG
    )
    def test_schema_fields_with_filter_fields_list(self):
        """Test schema field generator"""
        fields = self.backend.get_schema_fields(self.view)
        fields = [f.name for f in fields]
        self.assertEqual(fields, list(self.view.nested_filter_fields.keys()))

    @unittest.skipIf(
        not CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED, CORE_API_AND_CORE_SCHEMA_MISSING_MSG
    )
    def test_schema_field_not_required(self):
        """Test schema fields always not required"""
        fields = self.backend.get_schema_fields(self.view)
        fields = [f.required for f in fields]
        for field in fields:
            self.assertFalse(field)


if __name__ == "__main__":
    unittest.main()



"""
Search backend. Most likely to be deprecated soon.
"""

import operator
import warnings

from django_elasticsearch_dsl import fields
from elasticsearch_dsl.query import Q
from rest_framework.filters import BaseFilterBackend
from rest_framework.settings import api_settings
import six

from ..mixins import FilterBackendMixin
from ...compat import coreapi, coreschema

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.search.historical"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("SearchFilterBackend",)


class SearchFilterBackend(BaseFilterBackend, FilterBackendMixin):
    """Search filter backend for Elasticsearch.

    Example:

        >>> from django_elasticsearch_dsl_drf_alt.filter_backends import (
        >>>     SearchFilterBackend
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.viewsets import (
        >>>     BaseDocumentViewSet,
        >>> )
        >>>
        >>> # Local article document definition
        >>> from .documents import ArticleDocument
        >>>
        >>> # Local article document serializer
        >>> from .serializers import ArticleDocumentSerializer
        >>>
        >>> class ArticleDocumentView(BaseDocumentViewSet):
        >>>
        >>>     document = ArticleDocument
        >>>     serializer_class = ArticleDocumentSerializer
        >>>     filter_backends = [SearchFilterBackend,]
        >>>     search_fields = (
        >>>         'title',
        >>>         'content',
        >>>     )
        >>>     search_nested_fields = {
        >>>         'state': ['name'],
        >>>         'documents.author': ['title', 'description'],
        >>>     }
    """

    search_param = api_settings.SEARCH_PARAM

    def get_search_query_params(self, request):
        """Get search query params.

        :param request: Django REST framework request.
        :type request: rest_framework.request.Request
        :return: List of search query params.
        :rtype: list
        """
        query_params = request.query_params.copy()
        return query_params.getlist(self.search_param, [])

    def construct_nested_search(self, request, view):
        """Construct nested search.

        We have to deal with two types of structures:

        Type 1:

        >>> search_nested_fields = {
        >>>     'country': {
        >>>         'path': 'country',
        >>>         'fields': ['name'],
        >>>     },
        >>>     'city': {
        >>>         'path': 'country.city',
        >>>         'fields': ['name'],
        >>>     },
        >>> }

        Type 2:

        >>> search_nested_fields = {
        >>>     'country': {
        >>>         'path': 'country',
        >>>         'fields': [{'name': {'boost': 2}}]
        >>>     },
        >>>     'city': {
        >>>         'path': 'country.city',
        >>>         'fields': [{'name': {'boost': 2}}]
        >>>     },
        >>> }

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        if not hasattr(view, "search_nested_fields"):
            return []

        # TODO: Support query boosting

        query_params = self.get_search_query_params(request)
        __queries = []
        for search_term in query_params:
            for label, options in view.search_nested_fields.items():
                queries = []
                path = options.get("path")

                for _field in options.get("fields", []):

                    # In case if we deal with structure 2
                    if isinstance(_field, dict):
                        # TODO: take options (such as boost) into consideration
                        field = "{}.{}".format(path, _field["name"])
                    # In case if we deal with structure 1
                    else:
                        field = "{}.{}".format(path, _field)

                    field_kwargs = {field: search_term}

                    queries.append(Q("match", **field_kwargs))

                __queries.append(
                    Q(
                        "nested",
                        path=path,
                        query=six.moves.reduce(operator.or_, queries),
                    )
                )

        return __queries

    def construct_search(self, request, view):
        """Construct search.

        We have to deal with two types of structures:

        Type 1:

        >>> search_fields = (
        >>>     'title',
        >>>     'description',
        >>>     'summary',
        >>> )

        Type 2:

        >>> search_fields = {
        >>>     'title': {'boost': 2},
        >>>     'description': None,
        >>>     'summary': None,
        >>> }

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        query_params = self.get_search_query_params(request)
        __queries = []
        for search_term in query_params:
            __values = self.split_lookup_name(search_term, 1)
            __len_values = len(__values)
            if __len_values > 1:
                field, value = __values
                if field in view.search_fields:
                    # Initial kwargs for the match query
                    field_kwargs = {field: {"query": value}}
                    # In case if we deal with structure 2
                    if isinstance(view.search_fields, dict):
                        extra_field_kwargs = view.search_fields[field]
                        if extra_field_kwargs:
                            field_kwargs[field].update(extra_field_kwargs)
                    # The match query
                    __queries.append(Q("match", **field_kwargs))
            else:
                for field in view.search_fields:
                    # Initial kwargs for the match query
                    field_kwargs = {field: {"query": search_term}}

                    # In case if we deal with structure 2
                    if isinstance(view.search_fields, dict):
                        extra_field_kwargs = view.search_fields[field]
                        if extra_field_kwargs:
                            field_kwargs[field].update(extra_field_kwargs)

                    # The match query
                    __queries.append(Q("match", **field_kwargs))
        return __queries

    def filter_queryset(self, request, queryset, view):
        """Filter the queryset.

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        warnings.warn(
            "{} is deprecated. Switch to `CompoundSearchFilterBackend`."
            "".format(self.__class__.__name__)
        )
        __queries = self.construct_search(request, view) + self.construct_nested_search(
            request, view
        )

        if __queries:
            queryset = queryset.query("bool", should=__queries)
        return queryset

    def get_coreschema_field(self, field):
        if isinstance(field, fields.IntegerField):
            field_cls = coreschema.Number
        else:
            field_cls = coreschema.String
        return field_cls()

    def get_schema_fields(self, view):
        assert coreapi is not None, (
            "coreapi must be installed to " "use `get_schema_fields()`"
        )
        assert coreschema is not None, (
            "coreschema must be installed to " "use `get_schema_fields()`"
        )

        _search_fields = getattr(view, "search_fields", None)
        if isinstance(_search_fields, dict):
            search_fields = list(_search_fields.keys())
        else:
            search_fields = _search_fields

        return (
            []
            if not search_fields
            else [
                coreapi.Field(
                    name=self.search_param,
                    required=False,
                    location="query",
                    schema=coreschema.String(
                        description="Search in " "{}.".format(", ".join(search_fields))
                    ),
                )
            ]
        )



"""Compound search backend."""

from .base import BaseSearchFilterBackend
from .query_backends import (
    MatchQueryBackend,
    NestedQueryBackend,
)

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.search.compound"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("CompoundSearchFilterBackend",)


class CompoundSearchFilterBackend(BaseSearchFilterBackend):
    """Compound search backend."""

    query_backends = [
        MatchQueryBackend,
        NestedQueryBackend,
    ]



# coding: utf-8
"""
Test suggestions backend.
"""

from __future__ import absolute_import, unicode_literals

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

import factories

from ..versions import (
    ELASTICSEARCH_GTE_5_0,
    ELASTICSEARCH_GTE_6_0,
    ELASTICSEARCH_GTE_7_0,
)
from .base import BaseRestFrameworkTestCase
from .data_mixins import AddressesMixin

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_suggesters"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "TestContextSuggesters",
    "TestSuggesters",
    "TestSuggestersEmptyIndex",
)


@pytest.mark.django_db
class TestSuggesters(BaseRestFrameworkTestCase, AddressesMixin):
    """Test suggesters."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up class."""
        super(TestSuggesters, cls).setUpClass()

        cls.publishers = []
        cls.publishers.append(
            factories.PublisherFactory(
                name="Addison–Wesley",
                city="Brighton & Hove",
                state_province="East Midlands",
                country="Armenia",
            )
        )
        cls.publishers.append(
            factories.PublisherFactory(
                name="Adis International",
                city="Bristol",
                state_province="East of England",
                country="Argentina",
            )
        )
        cls.publishers.append(
            factories.PublisherFactory(
                name="Atlantic Books",
                city="Cardiff",
                state_province="North East",
                country="Belgium",
            )
        )
        cls.publishers.append(
            factories.PublisherFactory(
                name="Atlas Press",
                city="Carlisle",
                state_province="North West",
                country="Belarus",
            )
        )
        cls.publishers.append(
            factories.PublisherFactory(
                name="Book League of America",
                city="Chelmsford",
                state_province="South East",
                country="Burkina Faso",
            )
        )
        cls.publishers.append(
            factories.PublisherFactory(
                name="Book Works",
                city="Chester",
                state_province="South West",
                country="Burundi",
            )
        )
        cls.publishers.append(
            factories.PublisherFactory(
                name="Booktrope",
                city="Chichester",
                state_province="West Midlands",
                country="Netherlands",
            )
        )

        cls.publishers_url = reverse("publisherdocument-suggest", kwargs={})

        cls.books = []
        cls.books.append(
            factories.BookFactory(
                title="Aaaaa Bbbb",
                summary="`Twas brillig, and the slithy toves "
                "Did gyre and gimble in the wabe. "
                "All mimsy were the borogoves "
                "And the mome raths outgrabe.",
                publisher__name="Antares",
                publisher__country="Armenia",
            )
        )
        cls.books.append(
            factories.BookFactory(
                title="Aaaaa Cccc",
                summary='"Beware the Jabberwock, my son! '
                "The jaws that bite, the claws that catch! "
                "Beware the Jubjub bird, and shun "
                "The frumious Bandersnatch!",
                publisher__name="Antares",
                publisher__country="Armenia",
            )
        )
        cls.books.append(
            factories.BookFactory(
                title="Aaaaa Dddd",
                summary="He took his vorpal sword in his hand,"
                "Long time the manxome foe he sought --"
                "So rested he by the Tumtum tree,"
                "And stood awhile in thought.",
                publisher__name="Antares",
                publisher__country="Armenia",
            )
        )

        cls.books += factories.BookFactory.create_batch(
            10,
            publisher__name="Oxford University Press",
            publisher__city="Yerevan",
            publisher__state_province="Ararat",
            publisher__country="Ireland",
        )

        cls.books_url = reverse("bookdocument-suggest", kwargs={})

        cls.books_suggest_context_url = reverse(
            "bookdocument_frontend-suggest", kwargs={}
        )

        cls.authors = []
        cls.authors.append(
            factories.AuthorFactory(
                name="John Doe",
                salutation="Aaa Bbb",
            )
        )
        cls.authors.append(
            factories.AuthorFactory(
                name="Jane Doe",
                salutation="Aaa Ccc",
            )
        )
        cls.authors.append(
            factories.AuthorFactory(
                name="Armen Doe",
                salutation="Bbb Ccc",
            )
        )

        cls.authors_url = reverse("authordocument-suggest", kwargs={})

        cls.created_addresses()

        cls.sleep()
        call_command("search_index", "--rebuild", "-f")

    def _test_suggesters(self, test_data, url):
        """Test suggesters."""
        self.authenticate()

        data = {}

        for _suggester_field, _test_cases in test_data.items():

            for _test_case, _expected_results in _test_cases.items():
                _url = url + "?" + _suggester_field + "=" + _test_case
                # Check if response now is valid
                response = self.client.get(_url, data)
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertIn(_suggester_field, response.data)
                _unique_options = list(
                    set(
                        [
                            __o["text"]
                            for __o in response.data[_suggester_field][0]["options"]
                        ]
                    )
                )
                self.assertEqual(
                    len(_unique_options),
                    len(_expected_results),
                    (_url, _test_case, _expected_results, _unique_options),
                )

    def test_suggesters_completion(self):
        """Test suggesters completion."""
        # Testing publishers
        test_data = {
            "name_suggest__completion": {
                "Ad": ["Addison–Wesley", "Adis International"],
                "Atl": ["Atlantic Books", "Atlas Press"],
                "Boo": ["Book League of America", "Book Works", "Booktrope"],
            },
            "country_suggest__completion": {
                "Arm": ["Armenia"],
                "Ar": ["Armenia", "Argentina"],
                "Bel": ["Belgium", "Belarus"],
                "Bur": ["Burkina Faso", "Burundi"],
                "Net": ["Netherlands"],
                "Fra": [],
            },
        }
        # Testing default suggesters as well
        test_data.update(
            {
                "name_suggest": test_data["name_suggest__completion"],
                "country_suggest": test_data["country_suggest__completion"],
            }
        )
        self._test_suggesters(test_data, self.publishers_url)

        # Testing books
        test_data = {
            "title_suggest__completion": {
                "Aaa": ["Aaaaa Bbbb", "Aaaaa Cccc", "Aaaaa Dddd"],
                "Bbb": [],
            },
        }
        self._test_suggesters(test_data, self.books_url)

        # Testing authors
        test_data = {
            "salutation.suggest__completion": {
                "Aaa": ["Aaa Bbb", "Aaa Ccc"],
                "Bbb": ["Bbb Ccc"],
                "Hhh": [],
            },
        }
        self._test_suggesters(test_data, self.authors_url)

    def test_suggesters_completion_no_args_provided(self):
        """Test suggesters completion with no args provided."""
        data = {}
        # Check if response now is valid
        response = self.client.get(self.publishers_url, data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_suggesters_term(self):
        """Test suggesters term."""
        # Testing books
        test_data = {
            "summary_suggest__term": {
                "borogovse": ["borogov"],
                "Tumtus": ["tumtum"],
                "Jabberwok": ["jabberwock"],
                "tovse": ["tove", "took", "twas"],
            },
        }
        self._test_suggesters(test_data, self.books_url)

    def test_suggesters_phrase(self):
        """Test suggesters phrase."""
        # Testing books
        test_data = {
            "summary_suggest__phrase": {
                "slith tovs": ["slithi tov"],
                "mimsy boroto": ["mimsi borogov"],
            },
        }
        self._test_suggesters(test_data, self.books_url)

    def test_nested_fields_suggesters_completion(self):
        """Test suggesters completion for nested fields."""
        # Testing cities and countries
        test_data = {
            "city_suggest__completion": {
                "Ye": ["Yerevan", "Yeovil"],
                "Yer": ["Yerevan"],
                "Ams": ["Amsterdam"],
                "Du": ["Dublin"],
                "Ne": [],
            },
            "country_suggest__completion": {
                "Arm": ["Armenia"],
                "Ar": ["Armenia", "Argentina"],
                "Re": ["Republic of Ireland"],
                "Net": ["Netherlands"],
                "Fra": [],
            },
        }
        self._test_suggesters(test_data, self.addresses_suggest_url)


@pytest.mark.django_db
class TestSuggestersEmptyIndex(BaseRestFrameworkTestCase, AddressesMixin):
    """Test suggesters on empty index."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up class."""
        cls.authors_url = reverse("authordocument-suggest", kwargs={})

        cls.sleep()
        # Suggest on empty index
        call_command("search_index", "--delete", "-f")
        call_command("search_index", "--create", "-f")

    def test_suggesters_on_empty_index(self):
        """Test suggesters phrase."""
        response = self.client.get(
            self.authors_url + "?name_suggest__completion=Ad", {}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        if ELASTICSEARCH_GTE_7_0:
            self.assertTrue(bool(response.data.get("name_suggest__completion")))
        elif ELASTICSEARCH_GTE_6_0:
            self.assertFalse(bool(response.data))
        else:
            self.assertFalse(bool(response.data.get("name_suggest__completion")))


@unittest.skipIf(not ELASTICSEARCH_GTE_5_0, "ES >=5.x only")
@pytest.mark.django_db
class TestContextSuggesters(BaseRestFrameworkTestCase, AddressesMixin):
    """Test context suggesters."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up class."""
        # Books
        cls.books = []
        cls.books.append(
            factories.BookFactory(
                title="Ccccc Bbbb",
                summary="`Twas brillig, and the slithy toves "
                "Did gyre and gimble in the wabe. "
                "All mimsy were the borogoves "
                "And the mome raths outgrabe.",
                publisher__name="Antares",
                publisher__country="Armenia",
            )
        )
        cls.books.append(
            factories.BookFactory(
                title="Ccccc Cccc",
                summary='"Beware the Jabberwock, my son! '
                "The jaws that bite, the claws that catch! "
                "Beware the Jubjub bird, and shun "
                "The frumious Bandersnatch!",
                publisher__name="Antares",
                publisher__country="Armenia",
            )
        )
        cls.books.append(
            factories.BookFactory(
                title="Ccccc Dddd",
                summary="He took his vorpal sword in his hand,"
                "Long time the manxome foe he sought --"
                "So rested he by the Tumtum tree,"
                "And stood awhile in thought.",
                publisher__name="Antares",
                publisher__country="Armenia",
            )
        )
        cls.books.append(
            factories.BookFactory(
                title="Ccccc Eeee",
                summary="She took his zorpal blade in his hand,"
                "Long time the manxome foe he sought --"
                "So rested he by the Tumtum tree,"
                "And stood awhile in thought.",
                # publisher__name='Mario',
                # publisher__country='US',
            )
        )

        cls.books += factories.BookFactory.create_batch(
            10,
            publisher__name="Oxford University Press",
            publisher__city="Yerevan",
            publisher__state_province="Ararat",
            publisher__country="Ireland",
        )

        cls.books_suggest_context_url = reverse(
            "bookdocument_frontend-suggest", kwargs={}
        )

        # Addresses
        cls.addresses = []
        cls.addresses.append(
            factories.AddressFactory(
                street="Halabyan",
                city__name="Yerevan",
                latitude=40.0742719,
                longitude=44.1930605,
            )
        )
        cls.addresses.append(
            factories.AddressFactory(
                street="Hambardzumyan",
                city__name="Yerevan",
                latitude=40.01,
                longitude=44.01,
            )
        )
        cls.addresses.append(
            factories.AddressFactory(
                street="Haghartsin",
                city__name="Yerevan",
                latitude=39.92,
                longitude=43.92,
            )
        )
        cls.addresses.append(
            factories.AddressFactory(
                street="Hamazaspyan",
                city__name="Tatev",
                latitude=39.3793612,
                longitude=46.2480006,
            )
        )
        cls.addresses.append(
            factories.AddressFactory(
                street="Harazatyan",
                city__name="Tatev",
                latitude=39.3793612,
                longitude=46.2480006,
            )
        )
        cls.addresses.append(
            factories.AddressFactory(
                street="Hardewijk",
                city__name="Groningen",
                latitude=53.2246892,
                longitude=6.56429,
            )
        )
        cls.addresses.append(
            factories.AddressFactory(
                street="Haringstraat",
                city__name="Groningen",
                latitude=53.2246892,
                longitude=6.56429,
            )
        )

        cls.addresses_suggest_context_url = reverse(
            "addressdocument_frontend-suggest", kwargs={}
        )

        cls.sleep()
        call_command("search_index", "--rebuild", "-f")

    def _test_suggesters_completion_context(self, test_data, url):
        """Test suggesters completion context."""
        self.authenticate()

        data = {}

        for _suggester_field, _test_cases in test_data.items():

            for _test_case, _test_data in _test_cases.items():
                _url = url + "?" + _suggester_field + "=" + _test_case
                for _query_param, _value in _test_data["filters"].items():
                    _url += "&{}={}".format(_query_param, _value)
                # Check if response now is valid
                response = self.client.get(_url, data)
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertIn(_suggester_field, response.data)
                _unique_options = list(
                    set(
                        [
                            __o["text"]
                            for __o in response.data[_suggester_field][0]["options"]
                        ]
                    )
                )
                self.assertEqual(
                    len(_unique_options),
                    len(_test_data["expected_results"]),
                    (_url, _test_case, _test_data["expected_results"], _unique_options),
                )

    def test_suggesters_completion_context(self):
        """Test suggesters completion context."""
        # Testing books
        test_data = {
            "title_suggest_context": {
                "Ccc": {
                    "expected_results": [
                        "Ccccc Bbbb",
                        "Ccccc Cccc",
                        "Ccccc Dddd",
                    ],
                    "filters": {
                        "title_suggest_publisher": "Antares",
                    },
                },
            },
        }
        self._test_suggesters_completion_context(
            test_data, self.books_suggest_context_url
        )

        # Testing addresses
        test_data = {
            "street_suggest_context": {
                "Ha": {
                    "expected_results": [
                        "Halabyan",
                        "Hambardzumyan",
                        "Haghartsin",
                        "Hamazaspyan",
                        "Harazatyan",
                    ],
                    "filters": {
                        "title_suggest_loc": (
                            "40__44__1000km" if ELASTICSEARCH_GTE_6_0 else "40__44"
                        ),
                    },
                },
            },
        }
        self._test_suggesters_completion_context(
            test_data, self.addresses_suggest_context_url
        )


if __name__ == "__main__":
    unittest.main()



"""
Test highlight backend.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

import factories

from .base import BaseRestFrameworkTestCase

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_highlight"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestHighlight",)


@pytest.mark.django_db
class TestHighlight(BaseRestFrameworkTestCase):
    """Test highlight."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        super(TestHighlight, cls).setUpClass()
        cls.books_count = 10
        cls.books = factories.BookFactory.create_batch(
            cls.books_count,
        )

        cls.special_books_count = 10
        cls.special_books = factories.BookFactory.create_batch(
            cls.books_count,
            title="Twenty Thousand Leagues Under the Sea",
            description="""
            The title refers to the distance traveled while under the sea and
            not to a depth, as twenty thousand leagues is over six times the
            diameter, and nearly twice the circumference of the Earth
            """,
            summary="""
            Margaret Drabble argues that Twenty Thousand Leagues Under the Sea
            anticipated the ecology movement and shaped the French avant-garde
            """,
        )
        cls.all_books_count = cls.special_books_count + cls.books_count

        cls.sleep()
        call_command("search_index", "--rebuild", "-f")

    def _list_results_with_highlights(self):
        """List results with facets."""
        self.authenticate()

        url = reverse("bookdocument-list", kwargs={}) + "?search=twenty"
        all_highlights_url = url + "&highlight=summary&highlight=description"

        # Make request
        no_args_response = self.client.get(url, {})
        self.assertEqual(no_args_response.status_code, status.HTTP_200_OK)

        # Should contain 10 results
        self.assertEqual(
            len(no_args_response.data["results"]), self.special_books_count
        )

        for result in no_args_response.data["results"]:
            self.assertEqual(list(result["highlight"].keys()), ["title"])

        # Make request
        all_highlights_response = self.client.get(all_highlights_url, {})
        self.assertEqual(all_highlights_response.status_code, status.HTTP_200_OK)

        # Should contain 20 results
        self.assertEqual(
            len(all_highlights_response.data["results"]),
            self.special_books_count,
        )
        for result in all_highlights_response.data["results"]:
            self.assertEqual(
                set(["title", "description", "summary"]),
                set(result["highlight"].keys()),
            )

    def test_list_results_with_highlights(self):
        """Test list results with facets."""
        return self._list_results_with_highlights()


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Test helpers.
"""

from __future__ import absolute_import, unicode_literals

import unittest

from django.core.management import call_command

import pytest

from ..elasticsearch_helpers import delete_all_indices, get_all_indices

from .base import BaseTestCase

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_management_commands"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestManagementCommands",)


@pytest.mark.django_db
class TestManagementCommands(BaseTestCase):
    """Test management commands."""

    pytestmark = pytest.mark.django_db

    def _get_all_indices(self):
        """Get all indices."""
        delete_all_indices(with_protected=True)

        self.sleep()
        call_command("search_index", "--rebuild", "-f")

        res = set(get_all_indices())
        expected = {
            "test_address",
            "test_location",
            "test_publisher",
            "test_journal",
            "test_city",
            "test_author",
            "test_book",
            "test_tag",
        }

        self.assertSetEqual(res, expected)

    def _delete_all_indices(self):
        """Delete all indices."""
        call_command("elasticsearch_remove_indexes", "--with-protected")

        res = get_all_indices()

        self.assertEqual(res, [])

    def test_all(self):
        """Filter by field."""
        self._get_all_indices()
        self._delete_all_indices()


if __name__ == "__main__":
    unittest.main()



"""
Base tests.
"""

from time import sleep
import logging

from django.test import TransactionTestCase
import pytest
from rest_framework.test import APIClient

import factories

from ..pip_helpers import get_installed_packages

__title__ = "django_elasticsearch_dsl_drf_alt.tests.base"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "BaseRestFrameworkTestCase",
    "BaseTestCase",
    "CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED",
    "CORE_API_AND_CORE_SCHEMA_MISSING_MSG",
    "CORE_API_IS_INSTALLED",
    "CORE_SCHEMA_IS_INSTALLED",
    "INSTALLED_PACKAGES",
    "SleepMixin",
)

LOGGER = logging.getLogger(__name__)
INSTALLED_PACKAGES = get_installed_packages()
CORE_API_IS_INSTALLED = "coreapi" in INSTALLED_PACKAGES
CORE_SCHEMA_IS_INSTALLED = "coreschema" in INSTALLED_PACKAGES
CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED = (
    CORE_API_IS_INSTALLED and CORE_SCHEMA_IS_INSTALLED
)
CORE_API_AND_CORE_SCHEMA_MISSING_MSG = (
    "Skipped because coreapi or " "coreschema are not installed!"
)


WAIT_FOR_INDEX = 2


class SleepMixin(object):
    """Sleep mix-in."""

    @classmethod
    def sleep(cls, wait_for_index=WAIT_FOR_INDEX):
        """Sleep for N seconds.

        :param wait_for_index:
        :return:
        """
        sleep(wait_for_index)


@pytest.mark.django_db
class BaseRestFrameworkTestCase(TransactionTestCase, SleepMixin):
    """Base REST framework test case."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up class."""

        cls.client = APIClient()

        # Create test user.
        cls.user = factories.TestUsernameSuperAdminUserFactory()

    def authenticate(self):
        """Helper for logging in Genre Coordinator user.

        :return:
        """
        self.client.login(
            username=factories.auth_user.TEST_USERNAME,
            password=factories.auth_user.TEST_PASSWORD,
        )


@pytest.mark.django_db
class BaseTestCase(TransactionTestCase, SleepMixin):
    """Base test case."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up class."""

        cls.client = APIClient()

        # Create test user.
        cls.user = factories.TestUsernameSuperAdminUserFactory()
